# Copyright (C) 2003  CAMP
# Please see the accompanying LICENSE file for further information.
from math import pi, sqrt

import numpy as np
from numpy.linalg import solve, inv
from scipy.linalg import eigh

from gpaw.setup_data import SetupData
from gpaw.atom.configurations import configurations
from gpaw import __version__ as version
from gpaw.atom.all_electron import AllElectron, shoot
from gpaw.utilities import hartree
from gpaw.xc.hybrid import constructX, atomic_exact_exchange
from gpaw.atom.filter import Filter


class Generator(AllElectron):
    def __init__(self, symbol, xcname='LDA', scalarrel=True, corehole=None,
                 configuration=None,
                 nofiles=True, txt='-', gpernode=150,
                 orbital_free=False, tw_coeff=1.):
        AllElectron.__init__(self, symbol, xcname, scalarrel, corehole,
                             configuration, nofiles, txt, gpernode,
                             orbital_free, tw_coeff)

    def run(self, core='', rcut=1.0, extra=None,
            logderiv=False, vbar=None, exx=False, name=None,
            normconserving='', filter=(0.4, 1.75), rcutcomp=None,
            write_xml=True,
            empty_states='', yukawa_gamma=0.0):

        self.name = name

        self.core = core
        if isinstance(rcut, float):
            rcut_l = [rcut]
        else:
            rcut_l = rcut
        rcutmax = max(rcut_l)
        rcutmin = min(rcut_l)
        self.rcut_l = rcut_l

        if rcutcomp is None:
            rcutcomp = rcutmin
        self.rcutcomp = rcutcomp

        hfilter, xfilter = filter

        Z = self.Z

        n_j = self.n_j
        l_j = self.l_j
        f_j = self.f_j
        e_j = self.e_j

        if vbar is None:
            vbar = ('poly', rcutmin * 0.9)
        vbar_type, rcutvbar = vbar

        normconserving_l = [x in normconserving for x in 'spdf']

        # Parse core string:
        j = 0
        if core.startswith('['):
            a, core = core.split(']')
            core_symbol = a[1:]
            j = len(configurations[core_symbol][1])

        while core != '':
            assert n_j[j] == int(core[0])
            assert l_j[j] == 'spdf'.find(core[1])
            if j != self.jcorehole:
                assert f_j[j] == 2 * (2 * l_j[j] + 1)
            j += 1
            core = core[2:]

        njcore = j
        self.njcore = njcore

        while empty_states != '':
            n = int(empty_states[0])
            l = 'spdf'.find(empty_states[1])
            assert n == 1 + l + l_j.count(l)
            n_j.append(n)
            l_j.append(l)
            f_j.append(0.0)
            e_j.append(-0.01)
            empty_states = empty_states[2:]

        if 2 in l_j[njcore:]:
            # We have a bound valence d-state.  Add bound s- and
            # p-states if not already there:
            for l in [0, 1]:
                if l not in l_j[njcore:]:
                    n_j.append(1 + l + l_j.count(l))
                    l_j.append(l)
                    f_j.append(0.0)
                    e_j.append(-0.01)

        if l_j[njcore:] == [0] and Z > 2:
            # We have only a bound valence s-state and we are not
            # hydrogen and not helium.  Add bound p-state:
            n_j.append(n_j[njcore])
            l_j.append(1)
            f_j.append(0.0)
            e_j.append(-0.01)

        nj = len(n_j)

        self.Nv = sum(f_j[njcore:])
        self.Nc = sum(f_j[:njcore])

        lmaxocc = max(l_j[njcore:])
        lmax = max(l_j[njcore:])

        #  Parameters for orbital_free
        if self.orbital_free:
            self.n_j = [1]
            self.l_j = [0]
            self.f_j = [self.Z]
            self.e_j = [self.e_j[0]]

            n_j = self.n_j
            l_j = self.l_j
            f_j = self.f_j
            e_j = self.e_j
            nj = len(n_j)
            lmax = 0
            lmaxocc = 0

        # Do all-electron calculation:
        AllElectron.run(self)

        # Highest occupied atomic orbital:
        self.emax = max(e_j)

        N = self.N
        r = self.r
        dr = self.dr
        d2gdr2 = self.d2gdr2
        beta = self.beta

        dv = r**2 * dr

        t = self.text
        t()
        t('Generating PAW setup')
        if core != '':
            t('Frozen core:', core)

        # So far - no ghost-states:
        self.ghost = False

        # Calculate the kinetic energy of the core states:
        Ekincore = 0.0
        j = 0
        for f, e, u in zip(f_j[:njcore], e_j[:njcore], self.u_j[:njcore]):
            u = np.where(abs(u) < 1e-160, 0, u)  # XXX Numeric!
            k = e - np.sum((u**2 * self.vr * dr)[1:] / r[1:])
            Ekincore += f * k
            if j == self.jcorehole:
                self.Ekincorehole = k
            j += 1

        # Calculate core density:
        if njcore == 0:
            nc = np.zeros(N)
        else:
            uc_j = self.u_j[:njcore]
            uc_j = np.where(abs(uc_j) < 1e-160, 0, uc_j)  # XXX Numeric!
            nc = np.dot(f_j[:njcore], uc_j**2) / (4 * pi)
            nc[1:] /= r[1:]**2
            nc[0] = nc[1]

        self.nc = nc

        # Calculate core kinetic energy density
        if njcore == 0:
            tauc = np.zeros(N)
        else:
            tauc = self.radial_kinetic_energy_density(f_j[:njcore],
                                                      l_j[:njcore],
                                                      self.u_j[:njcore])
            t('Kinetic energy of the core from tauc =',
              np.dot(tauc * r * r, dr) * 4 * pi)

        # Order valence states with respect to angular momentum
        # quantum number:
        self.n_ln = n_ln = []
        self.f_ln = f_ln = []
        self.e_ln = e_ln = []
        for l in range(lmax + 1):
            n_n = []
            f_n = []
            e_n = []
            for j in range(njcore, nj):
                if l_j[j] == l:
                    n_n.append(n_j[j])
                    f_n.append(f_j[j])
                    e_n.append(e_j[j])
            n_ln.append(n_n)
            f_ln.append(f_n)
            e_ln.append(e_n)

        # Add extra projectors:
        if extra is not None:
            if len(extra) == 0:
                lmaxextra = 0
            else:
                lmaxextra = max(extra.keys())
            if lmaxextra > lmax:
                for l in range(lmax, lmaxextra):
                    n_ln.append([])
                    f_ln.append([])
                    e_ln.append([])
                lmax = lmaxextra
            for l in extra:
                nn = -1
                for e in extra[l]:
                    n_ln[l].append(nn)
                    f_ln[l].append(0.0)
                    e_ln[l].append(e)
                    nn -= 1
        else:
            # Automatic:

            # Make sure we have two projectors for each occupied channel:
            for l in range(lmaxocc + 1):
                if len(n_ln[l]) < 2 and not normconserving_l[l]:
                    # Only one - add one more:
                    assert len(n_ln[l]) == 1
                    n_ln[l].append(-1)
                    f_ln[l].append(0.0)
                    e_ln[l].append(1.0 + e_ln[l][0])

            if lmaxocc < 2 and lmaxocc == lmax:
                # Add extra projector for l = lmax + 1:
                n_ln.append([-1])
                f_ln.append([0.0])
                e_ln.append([0.0])
                lmax += 1

        self.lmax = lmax

        rcut_l.extend([rcutmin] * (lmax + 1 - len(rcut_l)))

        t('Cutoffs:')
        for rc, s in zip(rcut_l, 'spdf'):
            t('rc(%s)=%.3f' % (s, rc))
        t('rc(vbar)=%.3f' % rcutvbar)
        t('rc(comp)=%.3f' % rcutcomp)
        t('rc(nct)=%.3f' % rcutmax)
        t()
        t('Kinetic energy of the core states: %.6f' % Ekincore)

        # Allocate arrays:
        self.u_ln = u_ln = []  # phi * r
        self.s_ln = s_ln = []  # phi-tilde * r
        self.q_ln = q_ln = []  # p-tilde * r
        for l in range(lmax + 1):
            nn = len(n_ln[l])
            u_ln.append(np.zeros((nn, N)))
            s_ln.append(np.zeros((nn, N)))
            q_ln.append(np.zeros((nn, N)))

        # Fill in all-electron wave functions:
        for l in range(lmax + 1):
            # Collect all-electron wave functions:
            u_n = [self.u_j[j] for j in range(njcore, nj) if l_j[j] == l]
            for n, u in enumerate(u_n):
                u_ln[l][n] = u

        # Grid-index corresponding to rcut:
        gcut_l = [1 + int(rc * N / (rc + beta)) for rc in rcut_l]

        rcutfilter = xfilter * rcutmax
        self.rcutfilter = rcutfilter
        gcutfilter = 1 + int(rcutfilter * N / (rcutfilter + beta))
        gcutmax = 1 + int(rcutmax * N / (rcutmax + beta))

        # Outward integration of unbound states stops at 3 * rcut:
        gmax = int(3 * rcutmax * N / (3 * rcutmax + beta))
        assert gmax > gcutfilter

        # Calculate unbound extra states:
        c2 = -(r / dr)**2
        c10 = -d2gdr2 * r**2
        for l, (n_n, e_n, u_n) in enumerate(zip(n_ln, e_ln, u_ln)):
            for n, e, u in zip(n_n, e_n, u_n):
                if n < 0:
                    u[:] = 0.0
                    shoot(u, l, self.vr, e, self.r2dvdr, r, dr, c10, c2,
                          self.scalarrel, gmax=gmax)
                    u *= 1.0 / u[gcut_l[l]]

        charge = Z - self.Nv - self.Nc
        t('Charge: %.1f' % charge)
        t('Core electrons: %.1f' % self.Nc)
        t('Valence electrons: %.1f' % self.Nv)

        # Construct smooth wave functions:
        coefs = []
        for l, (u_n, s_n) in enumerate(zip(u_ln, s_ln)):
            nodeless = True
            gc = gcut_l[l]
            for u, s in zip(u_n, s_n):
                s[:] = u
                if normconserving_l[l]:
                    A = np.zeros((5, 5))
                    A[:4, 0] = 1.0
                    A[:4, 1] = r[gc - 2:gc + 2]**2
                    A[:4, 2] = A[:4, 1]**2
                    A[:4, 3] = A[:4, 1] * A[:4, 2]
                    A[:4, 4] = A[:4, 2]**2
                    A[4, 4] = 1.0
                    a = u[gc - 2:gc + 3] / r[gc - 2:gc + 3]**(l + 1)
                    a = np.log(a)

                    def f(x):
                        a[4] = x
                        b = solve(A, a)
                        r1 = r[:gc]
                        r2 = r1**2
                        rl1 = r1**(l + 1)
                        y = b[0] + r2 * (b[1] + r2 * (b[2] + r2 *
                                                      (b[3] + r2 * b[4])))
                        y = np.exp(y)
                        s[:gc] = rl1 * y
                        return np.dot(s**2, dr) - 1
                    x1 = 0.0
                    x2 = 0.001
                    f1 = f(x1)
                    f2 = f(x2)
                    while abs(f1) > 1e-6:
                        x0 = (x1 / f1 - x2 / f2) / (1 / f1 - 1 / f2)
                        f0 = f(x0)
                        if abs(f1) < abs(f2):
                            x2, f2 = x1, f1
                        x1, f1 = x0, f0

                else:
                    A = np.ones((4, 4))
                    A[:, 0] = 1.0
                    A[:, 1] = r[gc - 2:gc + 2]**2
                    A[:, 2] = A[:, 1]**2
                    A[:, 3] = A[:, 1] * A[:, 2]
                    a = u[gc - 2:gc + 2] / r[gc - 2:gc + 2]**(l + 1)
                    if 0:  # l < 2 and nodeless:
                        a = np.log(a)
                    a = solve(A, a)
                    r1 = r[:gc]
                    r2 = r1**2
                    rl1 = r1**(l + 1)
                    y = a[0] + r2 * (a[1] + r2 * (a[2] + r2 * (a[3])))
                    if 0:  # l < 2 and nodeless:
                        y = np.exp(y)
                    s[:gc] = rl1 * y

                coefs.append(a)
                if nodeless:
                    if not (s[1:gc] > 0.0).all():
                        raise RuntimeError(
                            'Error: The %d%s pseudo wave has a node!' %
                            (n_ln[l][0], 'spdf'[l]))
                    # Only the first state for each l must be nodeless:
                    nodeless = False

        # Calculate pseudo core density:
        gcutnc = 1 + int(rcutmax * N / (rcutmax + beta))
        self.nct = nct = nc.copy()
        A = np.ones((4, 4))
        A[0] = 1.0
        A[1] = r[gcutnc - 2:gcutnc + 2]**2
        A[2] = A[1]**2
        A[3] = A[1] * A[2]
        a = nc[gcutnc - 2:gcutnc + 2]
        a = solve(np.transpose(A), a)
        r2 = r[:gcutnc]**2
        nct[:gcutnc] = a[0] + r2 * (a[1] + r2 * (a[2] + r2 * a[3]))
        t('Pseudo-core charge: %.6f' % (4 * pi * np.dot(nct, dv)))

        # ... and the pseudo core kinetic energy density:
        tauct = tauc.copy()
        a = tauc[gcutnc - 2:gcutnc + 2]
        a = solve(np.transpose(A), a)
        tauct[:gcutnc] = a[0] + r2 * (a[1] + r2 * (a[2] + r2 * a[3]))

        # ... and the soft valence density:
        nt = np.zeros(N)
        for f_n, s_n in zip(f_ln, s_ln):
            nt += np.dot(f_n, s_n**2) / (4 * pi)
        nt[1:] /= r[1:]**2
        nt[0] = nt[1]
        nt += nct
        self.nt = nt

        # Calculate the shape function:
        x = r / rcutcomp
        gaussian = np.zeros(N)
        self.gamma = gamma = 10.0
        gaussian[:gmax] = np.exp(-gamma * x[:gmax]**2)
        gt = 4 * (gamma / rcutcomp**2)**1.5 / sqrt(pi) * gaussian
        t('Shape function alpha=%.3f' % (gamma / rcutcomp**2))
        norm = np.dot(gt, dv)
        #  print norm, norm-1
        assert abs(norm - 1) < 1e-2
        gt /= norm

        # Calculate smooth charge density:
        Nt = np.dot(nt, dv)
        rhot = nt - (Nt + charge / (4 * pi)) * gt
        t('Pseudo-electron charge', 4 * pi * Nt)

        vHt = np.zeros(N)
        hartree(0, rhot * r * dr, r, vHt)
        vHt[1:] /= r[1:]
        vHt[0] = vHt[1]

        vXCt = np.zeros(N)

        extra_xc_data = {}

        if self.xc.type != 'GLLB':
            self.xc.calculate_spherical(self.rgd,
                                        nt.reshape((1, -1)),
                                        vXCt.reshape((1, -1)))
        else:
            self.xc.get_smooth_xc_potential_and_energy_1d(vXCt)

            # Calculate extra-stuff for non-local functionals
            self.xc.get_extra_setup_data(extra_xc_data)

        vt = vHt + vXCt

        if self.orbital_free:
            vt /= self.tw_coeff

        # Construct zero potential:
        gc = 1 + int(rcutvbar * N / (rcutvbar + beta))
        if vbar_type == 'f':
            assert lmax == 2
            uf = np.zeros(N)
            l = 3

            # Solve for all-electron f-state:
            eps = 0.0
            shoot(uf, l, self.vr, eps, self.r2dvdr, r, dr, c10, c2,
                  self.scalarrel, gmax=gmax)
            uf *= 1.0 / uf[gc]

            # Fit smooth pseudo f-state polynomium:
            A = np.ones((4, 4))
            A[:, 0] = 1.0
            A[:, 1] = r[gc - 2:gc + 2]**2
            A[:, 2] = A[:, 1]**2
            A[:, 3] = A[:, 1] * A[:, 2]
            a = uf[gc - 2:gc + 2] / r[gc - 2:gc + 2]**(l + 1)
            a0, a1, a2, a3 = solve(A, a)
            r1 = r[:gc]
            r2 = r1**2
            rl1 = r1**(l + 1)
            y = a0 + r2 * (a1 + r2 * (a2 + r2 * a3))
            sf = uf.copy()
            sf[:gc] = rl1 * y

            # From 0 to gc, use analytic formula for kinetic energy operator:
            r4 = r2**2
            r6 = r4 * r2
            enumerator = (a0 * l * (l + 1) +
                          a1 * (l + 2) * (l + 3) * r2 +
                          a2 * (l + 4) * (l + 5) * r4 +
                          a3 * (l + 6) * (l + 7) * r6)
            denominator = a0 + a1 * r2 + a2 * r4 + a3 * r6
            ekin_over_phit = - 0.5 * (enumerator / denominator - l * (l + 1))
            ekin_over_phit[1:] /= r2[1:]

            vbar = eps - vt
            vbar[:gc] -= ekin_over_phit
            vbar[0] = vbar[1]  # Actually we can collect the terms into
            # a single fraction without poles, so as to avoid doing this,
            # but this is good enough

            # From gc to gmax, use finite-difference formula for kinetic
            # energy operator:
            vbar[gc:gmax] -= self.kin(l, sf)[gc:gmax] / sf[gc:gmax]
            vbar[gmax:] = 0.0
        else:
            assert vbar_type == 'poly'
            A = np.ones((2, 2))
            A[0] = 1.0
            A[1] = r[gc - 1:gc + 1]**2
            a = vt[gc - 1:gc + 1]
            a = solve(np.transpose(A), a)
            r2 = r**2
            vbar = a[0] + r2 * a[1] - vt
            vbar[gc:] = 0.0

        vt += vbar

        # Construct projector functions:
        for l, (e_n, s_n, q_n) in enumerate(zip(e_ln, s_ln, q_ln)):
            for e, s, q in zip(e_n, s_n, q_n):
                q[:] = self.kin(l, s) + (vt - e) * s
                q[gcutmax:] = 0.0

        filter = Filter(r, dr, gcutfilter, hfilter).filter
        if self.orbital_free:
            vbar *= self.tw_coeff
        vbar = filter(vbar * r)

        # Calculate matrix elements:
        self.dK_lnn = dK_lnn = []
        self.dH_lnn = dH_lnn = []
        self.dO_lnn = dO_lnn = []

        for l, (e_n, u_n, s_n, q_n) in enumerate(zip(e_ln, u_ln,
                                                     s_ln, q_ln)):
            A_nn = np.inner(s_n, q_n * dr)
            # Do a LU decomposition of A:
            nn = len(e_n)
            L_nn = np.identity(nn, float)
            U_nn = A_nn.copy()

            # Keep all bound states normalized
            if sum([n > 0 for n in n_ln[l]]) <= 1:
                for i in range(nn):
                    for j in range(i + 1, nn):
                        L_nn[j, i] = 1.0 * U_nn[j, i] / U_nn[i, i]
                        U_nn[j, :] -= U_nn[i, :] * L_nn[j, i]

            dO_nn = (np.inner(u_n, u_n * dr) -
                     np.inner(s_n, s_n * dr))

            e_nn = np.zeros((nn, nn))
            e_nn.ravel()[::nn + 1] = e_n
            dH_nn = np.dot(dO_nn, e_nn) - A_nn

            q_n[:] = np.dot(inv(np.transpose(U_nn)), q_n)
            s_n[:] = np.dot(inv(L_nn), s_n)
            u_n[:] = np.dot(inv(L_nn), u_n)

            dO_nn = np.dot(np.dot(inv(L_nn), dO_nn),
                           inv(np.transpose(L_nn)))
            dH_nn = np.dot(np.dot(inv(L_nn), dH_nn),
                           inv(np.transpose(L_nn)))

            ku_n = [self.kin(l, u, e) for u, e in zip(u_n, e_n)]
            ks_n = [self.kin(l, s) for s in s_n]
            dK_nn = 0.5 * (np.inner(u_n, ku_n * dr) -
                           np.inner(s_n, ks_n * dr))
            dK_nn += np.transpose(dK_nn).copy()

            dK_lnn.append(dK_nn)
            dO_lnn.append(dO_nn)
            dH_lnn.append(dH_nn)

            for n, q in enumerate(q_n):
                q[:] = filter(q, l) * r**(l + 1)

            A_nn = np.inner(s_n, q_n * dr)
            q_n[:] = np.dot(inv(np.transpose(A_nn)), q_n)

        self.vt = vt
        self.vbar = vbar

        t('state    eigenvalue         norm')
        t('--------------------------------')
        for l, (n_n, f_n, e_n) in enumerate(zip(n_ln, f_ln, e_ln)):
            for n in range(len(e_n)):
                if n_n[n] > 0:
                    f = '(%d)' % f_n[n]
                    t('%d%s%-4s: %12.6f %12.6f' % (
                        n_n[n], 'spdf'[l], f, e_n[n],
                        np.dot(s_ln[l][n]**2, dr)))
                else:
                    t('*%s    : %12.6f' % ('spdf'[l], e_n[n]))
        t('--------------------------------')

        self.logd = {}
        if logderiv:
            ni = 300
            self.elog = np.linspace(-5.0, 1.0, ni)
            # Calculate logarithmic derivatives:
            gld = gcutmax + 10
            self.rlog = r[gld]
            assert gld < gmax
            t('Calculating logarithmic derivatives at r=%.3f' % r[gld])
            t('(skip with [Ctrl-C])')

            try:
                u = np.zeros(N)
                for l in range(4):
                    self.logd[l] = (np.empty(ni), np.empty(ni))
                    if l <= lmax:
                        dO_nn = dO_lnn[l]
                        dH_nn = dH_lnn[l]
                        q_n = q_ln[l]

                    ae = self.symbol + '.ae.ld.' + 'spdf'[l]
                    ps = self.symbol + '.ps.ld.' + 'spdf'[l]
                    with open(ae, 'w') as fae, open(ps, 'w') as fps:
                        for i, e in enumerate(self.elog):
                            # All-electron logarithmic derivative:
                            u[:] = 0.0
                            shoot(u, l, self.vr, e, self.r2dvdr, r, dr,
                                  c10, c2, self.scalarrel, gmax=gld)
                            dudr = 0.5 * (u[gld + 1] - u[gld - 1]) / dr[gld]
                            ld = dudr / u[gld] - 1.0 / r[gld]
                            print(e, ld, file=fae)
                            self.logd[l][0][i] = ld

                            # PAW logarithmic derivative:
                            s = self.integrate(l, vt, e, gld)
                            if l <= lmax:
                                A_nn = dH_nn - e * dO_nn
                                s_n = [self.integrate(l, vt, e, gld, q)
                                       for q in q_n]
                                B_nn = np.inner(q_n, s_n * dr)
                                a_n = np.dot(q_n, s * dr)

                                B_nn = np.dot(A_nn, B_nn)
                                B_nn.ravel()[::len(a_n) + 1] += 1.0
                                c_n = solve(B_nn, np.dot(A_nn, a_n))
                                s -= np.dot(c_n, s_n)

                            dsdr = 0.5 * (s[gld + 1] - s[gld - 1]) / dr[gld]
                            ld = dsdr / s[gld] - 1.0 / r[gld]
                            print(e, ld, file=fps)
                            self.logd[l][1][i] = ld

            except KeyboardInterrupt:
                pass

        self.write(nc, 'nc')
        self.write(nt, 'nt')
        self.write(nct, 'nct')
        self.write(vbar, 'vbar')
        self.write(vt, 'vt')
        self.write(tauc, 'tauc')
        self.write(tauct, 'tauct')

        for l, (n_n, f_n, u_n, s_n, q_n) in enumerate(zip(n_ln, f_ln,
                                                          u_ln, s_ln, q_ln)):
            for n, f, u, s, q in zip(n_n, f_n, u_n, s_n, q_n):
                if n < 0:
                    self.write(u, 'ae', n=n, l=l)
                self.write(s, 'ps', n=n, l=l)
                self.write(q, 'proj', n=n, l=l)

        # Test for ghost states:
        for h in [0.05]:
            self.diagonalize(h)

        self.vn_j = vn_j = []
        self.vl_j = vl_j = []
        self.vf_j = vf_j = []
        self.ve_j = ve_j = []
        self.vu_j = vu_j = []
        self.vs_j = vs_j = []
        self.vq_j = vq_j = []
        j_ln = [[0 for f in f_n] for f_n in f_ln]
        j = 0
        for l, n_n in enumerate(n_ln):
            for n, nn in enumerate(n_n):
                if nn > 0:
                    vf_j.append(f_ln[l][n])
                    vn_j.append(nn)
                    vl_j.append(l)
                    ve_j.append(e_ln[l][n])
                    vu_j.append(u_ln[l][n])
                    vs_j.append(s_ln[l][n])
                    vq_j.append(q_ln[l][n])
                    j_ln[l][n] = j
                    j += 1
        for l, n_n in enumerate(n_ln):
            for n, nn in enumerate(n_n):
                if nn < 0:
                    vf_j.append(0)
                    vn_j.append(nn)
                    vl_j.append(l)
                    ve_j.append(e_ln[l][n])
                    vu_j.append(u_ln[l][n])
                    vs_j.append(s_ln[l][n])
                    vq_j.append(q_ln[l][n])
                    j_ln[l][n] = j
                    j += 1
        nj = j

        self.dK_jj = np.zeros((nj, nj))
        for l, j_n in enumerate(j_ln):
            for n1, j1 in enumerate(j_n):
                for n2, j2 in enumerate(j_n):
                    self.dK_jj[j1, j2] = self.dK_lnn[l][n1, n2]
                    if self.orbital_free:
                        self.dK_jj[j1, j2] *= self.tw_coeff

        X_gamma = yukawa_gamma
        if exx:
            X_p = constructX(self)
            if yukawa_gamma is not None and yukawa_gamma > 0:
                X_pg = constructX(self, yukawa_gamma)
            else:
                X_pg = None
            ExxC = atomic_exact_exchange(self, 'core-core')
        else:
            X_p = None
            X_pg = None
            ExxC = None

        sqrt4pi = sqrt(4 * pi)
        setup = SetupData(self.symbol, self.xc.name, self.name,
                          readxml=False,
                          generator_version=1)

        def divide_by_r(x_g, l):
            r = self.r
            # for x_g, l in zip(x_jg, l_j):
            p = x_g.copy()
            p[1:] /= self.r[1:]
            # XXXXX go to higher order!!!!!
            if l == 0:  # l_j[self.jcorehole] == 0:
                p[0] = (p[2] +
                        (p[1] - p[2]) * (r[0] - r[2]) / (r[1] - r[2]))
            return p

        def divide_all_by_r(x_jg):
            return [divide_by_r(x_g, l) for x_g, l in zip(x_jg, vl_j)]

        setup.l_j = vl_j
        setup.l_orb_J = vl_j
        setup.n_j = vn_j
        setup.f_j = vf_j
        setup.eps_j = ve_j
        setup.rcut_j = [rcut_l[l] for l in vl_j]

        setup.nc_g = nc * sqrt4pi
        setup.nct_g = nct * sqrt4pi
        setup.nvt_g = (nt - nct) * sqrt4pi
        setup.e_kinetic_core = Ekincore
        setup.vbar_g = vbar * sqrt4pi
        setup.tauc_g = tauc * sqrt4pi
        setup.tauct_g = tauct * sqrt4pi
        setup.extra_xc_data = extra_xc_data
        setup.Z = Z
        setup.Nc = self.Nc
        setup.Nv = self.Nv
        setup.e_kinetic = self.Ekin
        setup.e_xc = self.Exc
        setup.e_electrostatic = self.e_coulomb
        setup.e_total = self.e_coulomb + self.Exc + self.Ekin

        setup.rgd = self.rgd

        setup.shape_function = {'type': 'gauss',
                                'rc': self.rcutcomp / sqrt(self.gamma)}
        setup.e_kin_jj = self.dK_jj
        setup.ExxC = ExxC
        setup.phi_jg = divide_all_by_r(vu_j)
        setup.phit_jg = divide_all_by_r(vs_j)
        setup.pt_jg = divide_all_by_r(vq_j)
        setup.X_p = X_p
        setup.X_pg = X_pg
        setup.X_gamma = X_gamma

        if self.jcorehole is not None:
            setup.has_corehole = True
            setup.lcorehole = l_j[self.jcorehole]  # l_j or vl_j ????? XXX
            setup.ncorehole = n_j[self.jcorehole]
            setup.phicorehole_g = divide_by_r(self.u_j[self.jcorehole],
                                              setup.lcorehole)
            setup.core_hole_e = self.e_j[self.jcorehole]
            setup.core_hole_e_kin = self.Ekincorehole
            setup.fcorehole = self.fcorehole

        if self.ghost and not self.orbital_free:
            # In orbital_free we are not interested in ghosts
            raise RuntimeError('Ghost!')

        if self.scalarrel:
            reltype = 'scalar-relativistic'
        else:
            reltype = 'non-relativistic'

        attrs = [('type', reltype), ('name', 'gpaw-%s' % version)]
        data = 'Frozen core: ' + (self.core or 'none')

        setup.generatorattrs = attrs
        setup.generatordata = data
        setup.orbital_free = self.orbital_free
        setup.version = '0.6'

        self.id_j = []
        for l, n in zip(vl_j, vn_j):
            if n > 0:
                id = '%s-%d%s' % (self.symbol, n, 'spdf'[l])
            else:
                id = '%s-%s%d' % (self.symbol, 'spdf'[l], -n)
            self.id_j.append(id)
        setup.id_j = self.id_j

        if write_xml:
            setup.write_xml()
        return setup

    def diagonalize(self, h):
        ng = 350
        t = self.text
        t()
        t('Diagonalizing with gridspacing h=%.3f' % h)
        R = h * np.arange(1, ng + 1)
        G = (self.N * R / (self.beta + R) + 0.5).astype(int)
        G = np.clip(G, 1, self.N - 2)
        R1 = np.take(self.r, G - 1)
        R2 = np.take(self.r, G)
        R3 = np.take(self.r, G + 1)
        x1 = (R - R2) * (R - R3) / (R1 - R2) / (R1 - R3)
        x2 = (R - R1) * (R - R3) / (R2 - R1) / (R2 - R3)
        x3 = (R - R1) * (R - R2) / (R3 - R1) / (R3 - R2)

        def interpolate(f):
            f1 = np.take(f, G - 1)
            f2 = np.take(f, G)
            f3 = np.take(f, G + 1)
            return f1 * x1 + f2 * x2 + f3 * x3
        vt = interpolate(self.vt)
        t()
        t('state   all-electron     PAW')
        t('-------------------------------')
        for l in range(4):
            if l <= self.lmax:
                q_n = np.array([interpolate(q) for q in self.q_ln[l]])
                H = np.dot(np.transpose(q_n),
                           np.dot(self.dH_lnn[l], q_n)) * h
                S = np.dot(np.transpose(q_n),
                           np.dot(self.dO_lnn[l], q_n)) * h
            else:
                H = np.zeros((ng, ng))
                S = np.zeros((ng, ng))
            H.ravel()[::ng + 1] += vt + 1.0 / h**2 + l * (l + 1) / 2.0 / R**2
            H.ravel()[1::ng + 1] -= 0.5 / h**2
            H.ravel()[ng::ng + 1] -= 0.5 / h**2
            S.ravel()[::ng + 1] += 1.0
            e_n, _ = eigh(H, S)
            ePAW = e_n[0]
            if l <= self.lmax and self.n_ln[l][0] > 0:
                eAE = self.e_ln[l][0]
                t('%d%s:   %12.6f %12.6f' % (self.n_ln[l][0],
                                             'spdf'[l], eAE, ePAW), end='')
                if abs(eAE - ePAW) > 0.014:
                    t('  GHOST-STATE!')
                    self.ghost = True
                else:
                    t()
            else:
                t('*%s:                %12.6f' % ('spdf'[l], ePAW), end='')
                if ePAW < self.emax:
                    t('  GHOST-STATE!')
                    self.ghost = True
                else:
                    t()
        t('-------------------------------')

    def integrate(self, l, vt, e, gld, q=None):
        r = self.r[1:]
        dr = self.dr[1:]
        s = np.zeros(self.N)

        c0 = 0.5 * l * (l + 1) / r**2
        c1 = -0.5 * self.d2gdr2[1:]
        c2 = -0.5 * dr**-2

        fp = c2 + 0.5 * c1
        fm = c2 - 0.5 * c1
        f0 = c0 - 2 * c2

        f0 += vt[1:] - e
        if q is None:
            s[1] = r[1]**(l + 1)
            for g in range(gld):
                s[g + 2] = (-fm[g] * s[g] - f0[g] * s[g + 1]) / fp[g]
            return s

        s[1] = q[1] / (vt[0] - e)
        for g in range(gld):
            s[g + 2] = (q[g + 1] - fm[g] * s[g] - f0[g] * s[g + 1]) / fp[g]
        return s


def construct_smooth_wavefunction(u, l, gc, r, s):
    # Do a linear regression to a wave function
    # s = a + br^2 + cr^4 + dr^6, such that
    # the fitting is as good as possible in region gc-2:gc+2
    A = np.ones((4, 4))
    A[:, 0] = 1.0
    A[:, 1] = r[gc - 2:gc + 2]**2
    A[:, 2] = A[:, 1]**2
    A[:, 3] = A[:, 1] * A[:, 2]
    a = u[gc - 2:gc + 2] / r[gc - 2:gc + 2]**(l + 1)
    a = solve(A, a)
    r1 = r[:gc]
    r2 = r1**2
    rl1 = r1**(l + 1)
    y = a[0] + r2 * (a[1] + r2 * (a[2] + r2 * (a[3])))
    s[:gc] = rl1 * y


if __name__ == '__main__':
    import os
    from gpaw.atom.basis import BasisMaker
    from gpaw.atom.configurations import parameters

    for xcname in ['LDA', 'PBE', 'RPBE', 'revPBE', 'GLLBSC']:
        for symbol, par in parameters.items():
            filename = symbol + '.' + xcname
            if os.path.isfile(filename) or os.path.isfile(filename + '.gz'):
                continue
            g = Generator(symbol, xcname, scalarrel=True, nofiles=True)
            g.run(exx=True, logderiv=False, **par)

            if xcname == 'PBE':
                bm = BasisMaker(g, name='dzp', run=False)
                basis = bm.generate()
                basis.write_xml()



"""GPAW Response core functionality."""
from .groundstate import ResponseGroundStateAdapter  # noqa
from .context import ResponseContext, timer  # noqa

__all__ = ['ResponseGroundStateAdapter', 'ResponseContext', 'timer']



# Initially written by Lauri Lehtovaara, 2007
"""This module implements time propagators for time-dependent density
functional theory calculations."""
from abc import ABC, abstractmethod
from collections import namedtuple

import numpy as np

from ase.utils.timing import timer

from gpaw.utilities.blas import axpy

from gpaw.tddft.utils import MultiBlas
from gpaw.tddft.tdopers import DummyDensity


def create_propagator(name, **kwargs):
    if isinstance(name, BasePropagator):
        return name
    elif isinstance(name, dict):
        kwargs.update(name)
        return create_propagator(**kwargs)
    elif name == 'ECN':
        return ExplicitCrankNicolson(**kwargs)
    elif name == 'SICN':
        return SemiImplicitCrankNicolson(**kwargs)
    elif name == 'EFSICN':
        return EhrenfestPAWSICN(**kwargs)
    elif name == 'EFSICN_HGH':
        return EhrenfestHGHSICN(**kwargs)
    elif name == 'ETRSCN':
        return EnforcedTimeReversalSymmetryCrankNicolson(**kwargs)
    elif name == 'SITE':
        return SemiImplicitTaylorExponential(**kwargs)
    elif name == 'SIKE':
        return SemiImplicitKrylovExponential(**kwargs)
    elif name.startswith('SITE') or name.startswith('SIKE'):
        raise DeprecationWarning(
            'Use dictionary to specify degree.')
    else:
        raise ValueError('Unknown propagator: %s' % name)


def allocate_wavefunction_arrays(wfs):
    """Allocate wavefunction arrays."""
    DummyKPoint = namedtuple('DummyKPoint', ['psit_nG'])
    new_kpt_u = []
    for kpt in wfs.kpt_u:
        psit_nG = wfs.gd.empty(n=kpt.psit_nG.shape[0], dtype=complex)
        new_kpt_u.append(DummyKPoint(psit_nG))
    return new_kpt_u


class BasePropagator(ABC):
    """Abstract base class for time propagators."""

    def todict(self):
        return {'name': self.__class__.__name__}

    def initialize(self, td_density, td_hamiltonian, td_overlap, solver,
                   preconditioner, gd, timer):
        """Initialize propagator using runtime objects.

        Parameters
        ----------
        td_density: TimeDependentDensity
            the time-dependent density
        td_hamiltonian: TimeDependentHamiltonian
            the time-dependent hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (/wavefunction) grid descriptor
        timer: Timer
            timer
        """
        self.td_density = td_density
        self.td_hamiltonian = td_hamiltonian
        self.td_overlap = td_overlap

        self.wfs = td_density.get_wavefunctions()

        self.solver = solver
        self.preconditioner = preconditioner
        self.gd = gd
        self.timer = timer

        self.mblas = MultiBlas(gd)

    # Solve M psin = psi
    def apply_preconditioner(self, psi, psin):
        """Solves preconditioner equation.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result

        """
        self.timer.start('Solve TDDFT preconditioner')
        if self.preconditioner is not None:
            self.preconditioner.apply(self.kpt, psi, psin)
        else:
            psin[:] = psi
        self.timer.stop('Solve TDDFT preconditioner')

    @timer('Update time-dependent operators')
    def update_time_dependent_operators(self, time):
        """Update overlap, density, and Hamiltonian.

        Parameters
        ----------
        time: float
            the current time
        """
        # Update overlap S(t) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t) based on the wavefunctions psit_nG
        # in kpt_u for t = time. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Update Hamiltonian H(t) to reflect density rho(t)
        self.td_hamiltonian.update(self.td_density.get_density(), time)

    @timer('Update time-dependent operators')
    def half_update_time_dependent_operators(self, time):
        """Half-update overlap, density, and Hamiltonian.

        Parameters
        ----------
        time: float
            the time passed to hamiltonian.half_update()
        """
        # Update overlap S(t+dt) of kpt.psit_nG in kpt.P_ani.
        self.td_overlap.update(self.wfs)

        # Calculate density rho(t+dt) based on the wavefunctions psit_nG in
        # kpt_u for t = time+time_step. Updates wfs.D_asp based on kpt.P_ani.
        self.td_density.update()

        # Estimate Hamiltonian H(t+dt/2) by averaging H(t) and H(t+dt)
        # and retain the difference for a half-way Hamiltonian dH(t+dt/2).
        self.td_hamiltonian.half_update(self.td_density.get_density(), time)

        # Estimate overlap S(t+dt/2) by averaging S(t) and S(t+dt) #TODO!!!
        # XXX this doesn't do anything, see TimeDependentOverlap.half_update()
        self.td_overlap.half_update(self.wfs)

    @abstractmethod
    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            the time step

        """
        raise NotImplementedError()


class ExplicitCrankNicolson(BasePropagator):
    """Explicit Crank-Nicolson propagator

    Crank-Nicolson propagator, which approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)

    """
    def __init__(self):
        """Create ExplicitCrankNicolson-object."""
        BasePropagator.__init__(self)
        self.tmp_kpt_u = None
        self.hpsit = None
        self.spsit = None
        self.sinvhpsit = None

    def todict(self):
        return {'name': 'ECN'}

    def initialize(self, *args, **kwargs):
        BasePropagator.initialize(self, *args, **kwargs)

        # Allocate temporary wavefunctions
        self.tmp_kpt_u = allocate_wavefunction_arrays(self.wfs)

        # Allocate memory for Crank-Nicolson stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        self.hpsit = self.gd.zeros(nvec, dtype=complex)
        self.spsit = self.gd.zeros(nvec, dtype=complex)

    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def propagate(self, time, time_step):
        """Propagate wavefunctions.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step

        """
        self.niter = 0

        # Copy current wavefunctions psit_nG to work wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # Euler step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt,
                                            rhs_kpt,
                                            time_step,
                                            guess=True)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt,
                                                   rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt,
                                  rhs_kpt.psit_nG,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG,
                              self.spsit,
                              self.wfs,
                              kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        # rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j * time_step, self.hpsit, rhs_kpt.psit_nG,
                               nvec)

        if guess:
            if self.sinvhpsit is None:
                self.sinvhpsit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            # Update estimate of psit(t+dt) to ( 1 - i S^(-1) H dt ) psit(t)
            self.td_overlap.apply_inverse(self.hpsit,
                                          self.sinvhpsit,
                                          self.wfs,
                                          kpt,
                                          use_cg=False)
            self.mblas.multi_zaxpy(-1.0j * time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        self.timer.start('Apply time-dependent operators')
        self.td_overlap.update_k_point_projections(self.wfs, self.kpt, psi)
        self.td_hamiltonian.apply(self.kpt,
                                  psi,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(psi,
                              self.spsit,
                              self.wfs,
                              self.kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # psin[:] = self.spsit + .5J * self.time_step * self.hpsit
        psin[:] = self.spsit
        self.mblas.multi_zaxpy(.5j * self.time_step, self.hpsit, psin,
                               len(psi))


class SemiImplicitCrankNicolson(ExplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator

    Crank-Nicolson propagator, which first approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step to predict future
    wavefunctions. Then the approximations for the future wavefunctions are
    used to approximate the Hamiltonian at the middle of the time step.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)
    (S(t) + .5j dt H(t+dt/2) / hbar) psi(t+dt)
    = (S(t) - .5j dt H(t+dt/2) / hbar) psi(t)

    """
    def __init__(self):
        """Create SemiImplicitCrankNicolson-object."""
        ExplicitCrankNicolson.__init__(self)
        self.old_kpt_u = None

    def todict(self):
        return {'name': 'SICN'}

    def initialize(self, *args, **kwargs):
        ExplicitCrankNicolson.initialize(self, *args, **kwargs)

        # Allocate old wavefunctions
        self.old_kpt_u = allocate_wavefunction_arrays(self.wfs)

    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0
        nvec = len(self.wfs.kpt_u[0].psit_nG)

        # Copy current wavefunctions to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt,
                                            rhs_kpt,
                                            time_step,
                                            guess=True)

        self.half_update_time_dependent_operators(time + time_step)

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            # Average of psit(t) and predicted psit(t+dt)
            mean_psit_nG = 0.5 * (kpt.psit_nG + rhs_kpt.psit_nG)
            self.td_hamiltonian.half_apply(kpt, mean_psit_nG, self.hpsit)
            self.td_overlap.apply_inverse(self.hpsit,
                                          self.sinvhpsit,
                                          self.wfs,
                                          kpt,
                                          use_cg=False)

            # Update kpt.psit_nG to reflect
            # psit(t+dt) - i S^(-1) dH(t+dt/2) dt/2 psit(t+dt/2)
            kpt.psit_nG[:] = kpt.psit_nG - .5J * self.sinvhpsit * time_step
            self.mblas.multi_zaxpy(-.5j * time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter


class EhrenfestPAWSICN(SemiImplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator for Ehrenfest dynamics
       TODO: merge this with the ordinary SICN
    """
    def __init__(self,
                 corrector_guess=True,
                 predictor_guess=(True, False),
                 use_cg=(False, False)):
        """Create SemiImplicitCrankNicolson-object.

        Parameters
        ----------
        corrector_guess: Bool
            use initial guess for the corrector step (default is True)
        predictor_guess: (Bool, Bool)
            use (first, second) order initial guesses for the predictor step
            default is (True, False)
        use_cg: (Bool, Bool)
            use CG for calculating the inverse overlap (predictor, corrector)
            default is (False, False)

        """
        SemiImplicitCrankNicolson.__init__(self)
        self.old_kpt_u = None
        self.corrector_guess = corrector_guess
        self.predictor_guess = predictor_guess
        self.use_cg = use_cg

        # self.hsinvhpsit = None
        self.sinvh2psit = None

    def todict(self):
        return {'name': 'EFSICN'}

    def update_velocities(self, v_at_new, v_at_old=None):
        self.v_at = v_at_new.copy()
        if (v_at_old is not None):
            self.v_at_old = v_at_old.copy()

    def propagate(self, time, time_step, v_a):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0
        nvec = len(self.wfs.kpt_u[0].psit_nG)

        # update the atomic velocities which are required
        # for calculating the P term
        self.update_velocities(v_a)

        # Copy current wavefunctions to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        # self.v_at = self.v_at_old.copy() #v(t) for predictor step
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt,
                                            rhs_kpt,
                                            time_step,
                                            guess=self.predictor_guess[0])

        self.half_update_time_dependent_operators(time + time_step)

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            # Average of psit(t) and predicted psit(t+dt)
            if (self.corrector_guess):
                mean_psit_nG = 0.5 * (kpt.psit_nG + rhs_kpt.psit_nG)
                self.td_hamiltonian.half_apply(kpt, mean_psit_nG, self.hpsit)
                self.td_overlap.apply_inverse(self.hpsit,
                                              self.sinvhpsit,
                                              self.wfs,
                                              kpt,
                                              use_cg=self.use_cg[1])

                # Update kpt.psit_nG to reflect
                # psit(t+dt) - i S^(-1) dH(t+dt/2) dt/2 psit(t+dt/2)
                kpt.psit_nG[:] = kpt.psit_nG - .5J * self.sinvhpsit * time_step
                self.mblas.multi_zaxpy(-.5j * time_step, self.sinvhpsit,
                                       kpt.psit_nG, nvec)

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self,
                                   kpt,
                                   rhs_kpt,
                                   time_step,
                                   calculate_P_ani=False,
                                   guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt,
                                                   rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt,
                                  rhs_kpt.psit_nG,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_hamiltonian.calculate_paw_correction(rhs_kpt.psit_nG,
                                                     self.hpsit,
                                                     self.wfs,
                                                     kpt,
                                                     self.v_at,
                                                     calculate_P_ani=False)

        self.td_overlap.apply(rhs_kpt.psit_nG,
                              self.spsit,
                              self.wfs,
                              kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        # rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j * time_step, self.hpsit, rhs_kpt.psit_nG,
                               nvec)

        if guess:
            if self.sinvhpsit is None:
                self.sinvhpsit = self.gd.zeros(len(kpt.psit_nG), dtype=complex)

            if self.predictor_guess[1]:
                if self.sinvh2psit is None:
                    self.sinvh2psit = self.gd.zeros(len(kpt.psit_nG),
                                                    dtype=complex)

            # Update estimate of psit(t+dt) to ( 1 - i S^(-1) H dt ) psit(t)
            self.td_overlap.apply_inverse(self.hpsit,
                                          self.sinvhpsit,
                                          self.wfs,
                                          kpt,
                                          use_cg=self.use_cg[0])

            self.mblas.multi_zaxpy(-1.0j * time_step, self.sinvhpsit,
                                   kpt.psit_nG, nvec)
            if (self.predictor_guess[1]):
                self.td_hamiltonian.apply(kpt,
                                          self.sinvhpsit,
                                          self.sinvh2psit,
                                          calculate_P_ani=False)
                self.td_hamiltonian.calculate_paw_correction(
                    self.sinvhpsit,
                    self.sinvh2psit,
                    self.wfs,
                    kpt,
                    self.v_at,
                    calculate_P_ani=False)
                self.td_overlap.apply_inverse(self.sinvh2psit,
                                              self.sinvh2psit,
                                              self.wfs,
                                              kpt,
                                              use_cg=self.use_cg[0])
                self.mblas.multi_zaxpy(-.5 * time_step * time_step,
                                       self.sinvh2psit, kpt.psit_nG, nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)

    # ( S + i H dt/2 ) psi
    def dot(self, psi, psin):
        """Applies the propagator matrix to the given wavefunctions.

        Parameters
        ----------
        psi: List of coarse grids
            the known wavefunctions
        psin: List of coarse grids
            the result ( S + i H dt/2 ) psi

        """
        self.timer.start('Apply time-dependent operators')
        self.td_overlap.update_k_point_projections(self.wfs, self.kpt, psi)
        self.td_hamiltonian.apply(self.kpt,
                                  psi,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_hamiltonian.calculate_paw_correction(psi,
                                                     self.hpsit,
                                                     self.wfs,
                                                     self.kpt,
                                                     self.v_at,
                                                     calculate_P_ani=False)
        self.td_overlap.apply(psi,
                              self.spsit,
                              self.wfs,
                              self.kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # psin[:] = self.spsit + .5J * self.time_step * self.hpsit
        psin[:] = self.spsit
        self.mblas.multi_zaxpy(.5j * self.time_step, self.hpsit, psin,
                               len(psi))


class EhrenfestHGHSICN(SemiImplicitCrankNicolson):
    """Semi-implicit Crank-Nicolson propagator for Ehrenfest dynamics
       using HGH pseudopotentials

    """
    def __init__(self):
        """Create SemiImplicitCrankNicolson-object."""
        SemiImplicitCrankNicolson.__init__(self)

    def todict(self):
        return {'name': 'EFSICN_HGH'}

    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Copy current wavefunctions to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # Predictor step
        # Overwrite psit_nG in tmp_kpt_u by (1 - i S^(-1)(t) H(t) dt) psit_nG
        # from corresponding kpt_u in a Euler step before predicting psit(t+dt)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.tmp_kpt_u):
            self.solve_propagation_equation(kpt,
                                            rhs_kpt,
                                            time_step,
                                            guess=False)

        self.half_update_time_dependent_operators(time + time_step)

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):

            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    # ( S + i H dt/2 ) psit(t+dt) = ( S - i H dt/2 ) psit(t)
    def solve_propagation_equation(self,
                                   kpt,
                                   rhs_kpt,
                                   time_step,
                                   calculate_P_ani=False,
                                   guess=False):

        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt,
                                                   rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt,
                                  rhs_kpt.psit_nG,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG,
                              self.spsit,
                              self.wfs,
                              kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        # rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j * time_step, self.hpsit, rhs_kpt.psit_nG,
                               nvec)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)


class EnforcedTimeReversalSymmetryCrankNicolson(SemiImplicitCrankNicolson):
    """Enforced time-reversal symmetry Crank-Nicolson propagator

    Crank-Nicolson propagator, which first approximates the time-dependent
    Hamiltonian to be unchanged during one iteration step to predict future
    wavefunctions. Then the approximations for the future wavefunctions are
    used to approximate the Hamiltonian in the future.

    (S(t) + .5j dt H(t) / hbar) psi(t+dt) = (S(t) - .5j dt H(t) / hbar) psi(t)
    (S(t) + .5j dt H(t+dt) / hbar) psi(t+dt)
    = (S(t) - .5j dt H(t) / hbar) psi(t)

    """
    def __init__(self):
        SemiImplicitCrankNicolson.__init__(self)

    def todict(self):
        return {'name': 'ETRSCN'}

    def propagate(self, time, time_step, update_callback=None):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # Copy current wavefunctions to work and old wavefunction arrays
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.old_kpt_u[u].psit_nG[:] = kpt.psit_nG
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # Predictor step
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.create_rhs(rhs_kpt, kpt, time_step)

        if update_callback is not None:
            update_callback()

        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        # XXX why here is full update and not half update?
        # (compare to other propagators)
        self.update_time_dependent_operators(time + time_step)

        # Corrector step
        # Use predicted psit_nG in kpt_u as an initial guess, whereas the old
        # wavefunction in old_kpt_u are used to calculate rhs based on psit(t)
        for [kpt, rhs_kpt] in zip(self.wfs.kpt_u, self.old_kpt_u):
            self.solve_propagation_equation(kpt, rhs_kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    def create_rhs(self, rhs_kpt, kpt, time_step):
        # kpt is guess, rhs_kpt is used to calculate rhs and is overwritten
        nvec = len(rhs_kpt.psit_nG)

        assert kpt != rhs_kpt, 'Data race condition detected'
        assert len(kpt.psit_nG) == nvec, 'Incompatible lhs/rhs vectors'

        self.timer.start('Apply time-dependent operators')
        # Store H psi(t) as hpsit and S psit(t) as spsit
        self.td_overlap.update_k_point_projections(self.wfs, kpt,
                                                   rhs_kpt.psit_nG)
        self.td_hamiltonian.apply(kpt,
                                  rhs_kpt.psit_nG,
                                  self.hpsit,
                                  calculate_P_ani=False)
        self.td_overlap.apply(rhs_kpt.psit_nG,
                              self.spsit,
                              self.wfs,
                              kpt,
                              calculate_P_ani=False)
        self.timer.stop('Apply time-dependent operators')

        # Update rhs_kpt.psit_nG to reflect ( S - i H dt/2 ) psit(t)
        # rhs_kpt.psit_nG[:] = self.spsit - .5J * self.hpsit * time_step
        rhs_kpt.psit_nG[:] = self.spsit
        self.mblas.multi_zaxpy(-.5j * time_step, self.hpsit, rhs_kpt.psit_nG,
                               nvec)

    # ( S + i H(t+dt) dt/2 ) psit(t+dt) = ( S - i H(t) dt/2 ) psit(t)
    # rhs_kpt = ( S - i H(t) dt/2 ) psit(t)
    def solve_propagation_equation(self, kpt, rhs_kpt, time_step, guess=False):
        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # Solve A x = b where A is (S + i H dt/2) and b = rhs_kpt.psit_nG
        self.niter += self.solver.solve(self, kpt.psit_nG, rhs_kpt.psit_nG)


class AbsorptionKick:
    """Absorption kick propagator

    Absorption kick propagator::

      (S(t) + .5j dt p.r / hbar) psi(0+) = (S(t) - .5j dt p.r / hbar) psi(0-)

    where ``|p| = (eps e / hbar)``, and eps is field strength, e is elementary
    charge.

    """
    def __init__(self, wfs, abs_kick_hamiltonian, td_overlap, solver,
                 preconditioner, gd, timer):
        """Create AbsorptionKick-object.

        Parameters
        ----------
        wfs: FDWaveFunctions
            time-independent grid-based wavefunctions
        abs_kick_hamiltonian: AbsorptionKickHamiltonian
            the absorption kick hamiltonian
        td_overlap: TimeDependentOverlap
            the time-dependent overlap operator
        solver: LinearSolver
            solver for linear equations
        preconditioner: Preconditioner
            preconditioner for linear equations
        gd: GridDescriptor
            coarse (wavefunction) grid descriptor
        timer: Timer
            timer

        """
        self.propagator = ExplicitCrankNicolson()
        self.propagator.initialize(DummyDensity(wfs),
                                   abs_kick_hamiltonian, td_overlap,
                                   solver, preconditioner, gd, timer)

    def kick(self):
        """Excite all possible frequencies.
        """
        for l in range(self.propagator.td_hamiltonian.iterations):
            self.propagator.propagate(0, 1.0)


class SemiImplicitTaylorExponential(BasePropagator):
    """Semi-implicit Taylor exponential propagator
    exp(-i S^-1 H t) = 1 - i S^-1 H t + (1/2) (-i S^-1 H t)^2 + ...

    """
    def __init__(self, degree=4):
        """Create SemiImplicitTaylorExponential-object.

        Parameters
        ----------
        degree: integer
            Degree of the Taylor polynomial (default is 4)

        """
        raise RuntimeError('SITE propagator is unstable')
        BasePropagator.__init__(self)
        self.degree = degree
        self.tmp_kpt_u = None
        self.psin = None
        self.hpsit = None

    def todict(self):
        return {'name': 'SITE',
                'degree': self.degree}

    def initialize(self, *args, **kwargs):
        BasePropagator.initialize(self, *args, **kwargs)

        # Allocate temporary wavefunctions
        self.tmp_kpt_u = allocate_wavefunction_arrays(self.wfs)

        # Allocate memory for Taylor exponential stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        self.psin = self.gd.zeros(nvec, dtype=complex)
        self.hpsit = self.gd.zeros(nvec, dtype=complex)

    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        # copy current wavefunctions to temporary variable
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # predict for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        self.half_update_time_dependent_operators(time + time_step)

        # propagate psit(t), not psit(t+dt), in correct
        for u, kpt in enumerate(self.wfs.kpt_u):
            kpt.psit_nG[:] = self.tmp_kpt_u[u].psit_nG

        # correct for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    # psi(t) = exp(-i t S^-1 H) psi(0)
    # psi(t) = 1  + (-i S^-1 H t) (1 + (1/2) (-i S^-1 H t) (1 + ... ) )
    def solve_propagation_equation(self, kpt, time_step):

        nvec = len(kpt.psit_nG)

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt
        self.time_step = time_step

        # psin = psi(0)
        self.psin[:] = kpt.psit_nG
        for k in range(self.degree, 0, -1):
            # psin = psi(0) + (1/k) (-i S^-1 H t) psin
            self.td_hamiltonian.apply(kpt, self.psin, self.hpsit)
            # S psin = H psin
            self.psin[:] = self.hpsit
            self.niter += self.solver.solve(self, self.psin, self.hpsit)
            # psin = psi(0) + (-it/k) S^-1 H psin
            self.mblas.multi_scale(-1.0j * time_step / k, self.psin, nvec)
            self.mblas.multi_zaxpy(1.0, kpt.psit_nG, self.psin, nvec)

        kpt.psit_nG[:] = self.psin

    def dot(self, psit, spsit):
        self.td_overlap.apply(psit, spsit, self.wfs, self.kpt)


class SemiImplicitKrylovExponential(BasePropagator):
    """Semi-implicit Krylov exponential propagator


    """
    def __init__(self, degree=4):
        """Create SemiImplicitKrylovExponential-object.

        Parameters
        ----------
        degree: integer
            Degree of the Krylov subspace (default is 4)
        """
        BasePropagator.__init__(self)
        self.degree = degree
        self.kdim = degree + 1
        self.tmp_kpt_u = None
        self.lm = None
        self.em = None
        self.hm = None
        self.sm = None
        self.xm = None
        self.qm = None
        self.Hqm = None
        self.Sqm = None
        self.rqm = None

    def todict(self):
        return {'name': 'SIKE',
                'degree': self.degree}

    def initialize(self, *args, **kwargs):
        BasePropagator.initialize(self, *args, **kwargs)

        # Allocate temporary wavefunctions
        self.tmp_kpt_u = allocate_wavefunction_arrays(self.wfs)

        # Allocate memory for Krylov subspace stuff
        nvec = len(self.wfs.kpt_u[0].psit_nG)
        self.em = np.zeros((nvec, self.kdim), dtype=float)
        self.lm = np.zeros((nvec, ), dtype=complex)
        self.hm = np.zeros((nvec, self.kdim, self.kdim), dtype=complex)
        self.sm = np.zeros((nvec, self.kdim, self.kdim), dtype=complex)
        self.xm = np.zeros((nvec, self.kdim, self.kdim), dtype=complex)
        self.qm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        self.Hqm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        self.Sqm = self.gd.zeros((self.kdim, nvec), dtype=complex)
        self.rqm = self.gd.zeros((nvec, ), dtype=complex)

    def propagate(self, time, time_step):
        """Propagate wavefunctions once.

        Parameters
        ----------
        time: float
            the current time
        time_step: float
            time step
        """

        self.niter = 0

        self.time_step = time_step

        # copy current wavefunctions to temporary variable
        for u, kpt in enumerate(self.wfs.kpt_u):
            self.tmp_kpt_u[u].psit_nG[:] = kpt.psit_nG

        # predict for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        self.half_update_time_dependent_operators(time + time_step)

        # propagate psit(t), not psit(t+dt), in correct
        for u, kpt in enumerate(self.wfs.kpt_u):
            kpt.psit_nG[:] = self.tmp_kpt_u[u].psit_nG

        # correct for each k-point
        for kpt in self.wfs.kpt_u:
            self.solve_propagation_equation(kpt, time_step)

        self.update_time_dependent_operators(time + time_step)

        return self.niter

    # psi(t) = exp(-i t S^-1 H) psi(0)
    def solve_propagation_equation(self, kpt, time_step):

        nvec = len(kpt.psit_nG)
        tmp = np.zeros((nvec, ), complex)
        xm_tmp = np.zeros((nvec, self.kdim), complex)

        qm = self.qm
        Hqm = self.Hqm
        Sqm = self.Sqm

        # Information needed by solver.solve -> self.dot
        self.kpt = kpt

        scale = self.create_krylov_subspace(kpt, self.td_hamiltonian,
                                            self.td_overlap, self.qm, self.Hqm,
                                            self.Sqm)

        # Calculate hm and sm
        for i in range(self.kdim):
            for j in range(self.kdim):
                self.mblas.multi_zdotc(tmp, qm[i], Hqm[j], nvec)
                tmp *= self.gd.dv
                for k in range(nvec):
                    self.hm[k][i][j] = tmp[k]
                self.mblas.multi_zdotc(tmp, qm[i], Sqm[j], nvec)
                tmp *= self.gd.dv
                for k in range(nvec):
                    self.sm[k][i][j] = tmp[k]

        # Diagonalize
        # Propagate
        # psi(t) = Qm Xm exp(-i Em t) Xm^H Sm e_1
        #        = Qm Xm exp(-i Em t) Sm Qm^H S psi(0) ???
        #        = Qm Xm exp(-i Em t) y
        #        = Qm Xm z
        # y = Sm Qm^H S psi(0) = Xm^H Sm e_1
        # if Sm = I then y is the first row of Xm^*
        # and z = exp(-i Em t) y
        for k in range(nvec):
            (self.em[k], self.xm[k]) = np.linalg.eigh(self.hm[k])

        self.em = np.exp(self.em * (-1.0j * time_step))
        for k in range(nvec):
            z = self.em[k] * np.conj(self.xm[k, 0])
            xm_tmp[k][:] = np.dot(self.xm[k], z)
        kpt.psit_nG[:] = 0.0
        for k in range(nvec):
            for i in range(self.kdim):
                axpy(xm_tmp[k][i] / scale[k], self.qm[i][k], kpt.psit_nG[k])

    # Create Krylov subspace
    #    K_v = { psi, S^-1 H psi, (S^-1 H)^2 psi, ... }

    def create_krylov_subspace(self, kpt, h, s, qm, Hqm, Sqm):
        nvec = len(kpt.psit_nG)
        tmp = np.zeros((nvec, ), complex)
        scale = np.zeros((nvec, ), complex)
        scale[:] = 0.0
        rqm = self.rqm

        # q_0 = psi
        rqm[:] = kpt.psit_nG

        for i in range(self.kdim):
            qm[i][:] = rqm

            # S orthogonalize
            # q_i = q_i - sum_j<i <q_j|S|q_i> q_j
            for j in range(i):
                self.mblas.multi_zdotc(tmp, qm[i], Sqm[j], nvec)
                tmp *= self.gd.dv
                tmp = np.conj(tmp)
                self.mblas.multi_zaxpy(-tmp, qm[j], qm[i], nvec)

            # S q_i
            s.apply(qm[i], Sqm[i], self.wfs, kpt)
            self.mblas.multi_zdotc(tmp, qm[i], Sqm[i], nvec)
            tmp *= self.gd.dv
            self.mblas.multi_scale(1. / np.sqrt(tmp), qm[i], nvec)
            self.mblas.multi_scale(1. / np.sqrt(tmp), Sqm[i], nvec)
            if i == 0:
                scale[:] = 1 / np.sqrt(tmp)

            # H q_i
            h.apply(kpt, qm[i], Hqm[i])

            # S r = H q_i, (if stuff, to save one inversion)
            if i + 1 < self.kdim:
                rqm[:] = Hqm[i]
                self.solver.solve(self, rqm, Hqm[i])

        return scale

    def dot(self, psit, spsit):
        self.td_overlap.apply(psit, spsit, self.wfs, self.kpt)

    # Below this, just for testing & debug
    def Sdot(self, psit, spsit):
        self.apply_preconditioner(psit, self.tmp)
        self.td_overlap.apply(self.tmp, spsit, self.wfs, self.kpt)

    def Hdot(self, psit, spsit):
        self.apply_preconditioner(psit, self.tmp)
        self.td_hamiltonian.apply(self.kpt, self.tmp, spsit)

    def inverse_overlap(self, kpt_u, degree):
        self.dot = self.Sdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(10):
            self.solver.solve(self, self.kpt.psit_nG, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG,
                                   nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1 / np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_overlap.apply(self.kpt.psit_nG, self.tmp, self.wfs, self.kpt)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('S min eig = ', nrm2)

    def overlap(self, kpt_u, degree):
        self.dot = self.Sdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(100):
            self.tmp[:] = self.kpt.psit_nG
            self.td_overlap.apply(self.tmp, self.kpt.psit_nG, self.wfs,
                                  self.kpt)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG,
                                   nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1 / np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_overlap.apply(self.kpt.psit_nG, self.tmp, self.wfs, self.kpt)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('S max eig = ', nrm2)

    def inverse_hamiltonian(self, kpt_u, degree):
        self.dot = self.Hdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(10):
            self.solver.solve(self, self.kpt.psit_nG, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG,
                                   nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1 / np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_hamiltonian.apply(self.kpt, self.kpt.psit_nG, self.tmp)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('H min eig = ', nrm2)

    def hamiltonian(self, kpt_u, degree):
        self.dot = self.Hdot
        self.kpt = kpt_u[0]
        nvec = len(self.kpt.psit_nG)
        nrm2 = np.zeros(nvec, dtype=complex)
        self.tmp = self.gd.zeros(n=nvec, dtype=complex)

        for i in range(100):
            self.tmp[:] = self.kpt.psit_nG
            self.td_hamiltonian.apply(self.kpt, self.tmp, self.kpt.psit_nG)
            self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.kpt.psit_nG,
                                   nvec)
            nrm2 *= self.gd.dv
            self.mblas.multi_scale(1 / np.sqrt(nrm2), self.kpt.psit_nG, nvec)
        self.td_hamiltonian.apply(self.kpt, self.kpt.psit_nG, self.tmp)
        self.mblas.multi_zdotc(nrm2, self.kpt.psit_nG, self.tmp, nvec)
        nrm2 *= self.gd.dv
        print('H max eig = ', nrm2)



import numpy as np

from gpaw.rotation import rotation
from gpaw.utilities import pack, unpack2


class SingleAtomRotations:
    def __init__(self, R_sii):
        self.R_sii = R_sii

    @classmethod
    def new(cls, ni, l_j, R_slmm):
        nsym = len(R_slmm)
        R_sii = np.zeros((nsym, ni, ni))
        i1 = 0
        for l in l_j:
            i2 = i1 + 2 * l + 1
            for s, R_lmm in enumerate(R_slmm):
                R_sii[s, i1:i2, i1:i2] = R_lmm[l]
            i1 = i2
        return cls(R_sii)

    def symmetrize(self, a, D_aii, map_sa):
        ni = self.R_sii.shape[1]
        D_ii = np.zeros((ni, ni))
        for s, R_ii in enumerate(self.R_sii):
            D_ii += R_ii @ D_aii[map_sa[s][a]] @ R_ii.T
        return D_ii / len(map_sa)


class AtomRotations:
    def __init__(self, setups, id_a, symmetry):
        R_slmm = []
        for op_cc in symmetry.op_scc:
            op_vv = np.linalg.inv(symmetry.cell_cv) @ op_cc @ symmetry.cell_cv
            R_slmm.append([rotation(l, op_vv) for l in range(4)])

        rotations = {}
        for key, setup in setups.items():
            rotations[key] = SingleAtomRotations.new(setup.ni, setup.l_j,
                                                     R_slmm)

        self._rotations = rotations
        self._id_a = id_a

    def get_R_asii(self):
        return [self.get_by_a(a).R_sii for a in range(len(self._id_a))]

    def get_by_a(self, a):
        return self._rotations[self._id_a[a]]

    def symmetrize_atomic_density_matrices(self, D_asp, a_sa):
        if not D_asp:
            return

        nspins = next(iter(D_asp.values())).shape[0]

        for s in range(nspins):
            D_aii = [unpack2(D_asp[a][s]) for a in range(len(D_asp))]
            for a, D_ii in enumerate(D_aii):
                D_asp[a][s] = pack(self.get_by_a(a).symmetrize(a, D_aii, a_sa))



import numpy as np
import pytest
from ase import Atoms
from ase.calculators.test import numeric_force

import _gpaw
from gpaw import GPAW, Mixer, PoissonSolver
from gpaw.external import PointChargePotential
from gpaw.test import equal


@pytest.mark.later
def test_ext_potential_point_charge(in_tmp_dir):
    # Find coefs for polynomial:
    c = np.linalg.solve([[1, 1, 1, 1],
                         [0, 2, 4, 6],
                         [0, 2, 12, 30],
                         [1 / 3, 1 / 5, 1 / 7, 1 / 9]],
                        [1, -1, 2, 0.5])
    print(c)
    print(c * 32)
    x = np.linspace(0, 1, 101)
    v = np.polyval(c[::-1], x**2)
    equal((v * x**2 - x).sum() / 100, 0, 1e-5)

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(x, v)
        x = np.linspace(0.2, 1.5, 101)
        plt.plot(x, 1 / x)
        plt.show()

    # Low-level test:
    h = 0.1
    q = 2.2
    beg_v = np.zeros(3, int)
    h_v = np.ones(3) * h
    q_p = np.ones(1) * q
    R_pv = np.array([[0.1, 0.2, -0.3]])
    vext_G = np.zeros((1, 1, 1))
    rhot_G = np.ones((1, 1, 1))

    def f(rc):
        vext_G[:] = 0.0
        _gpaw.pc_potential(beg_v, h_v, q_p, R_pv, rc,
                           np.inf, 1.0, vext_G, None)
        return -vext_G[0, 0, 0]

    d = (R_pv[0]**2).sum()**0.5

    for rc in [0.9 * d, 1.1 * d]:
        if d > rc:
            v0 = q / d
        else:
            v0 = np.polyval(c[::-1], (d / rc)**2) * q / rc

        equal(f(rc), v0, 1e-12)

        F_pv = np.zeros((1, 3))
        _gpaw.pc_potential(beg_v, h_v, q_p, R_pv, rc, np.inf, 1.0,
                           vext_G, None, rhot_G, F_pv)
        eps = 0.0001
        for v in range(3):
            R_pv[0, v] += eps
            ep = f(rc)
            R_pv[0, v] -= 2 * eps
            em = f(rc)
            R_pv[0, v] += eps
            F = -(ep - em) / (2 * eps) * h**3
            equal(F, -F_pv[0, v], 1e-9)

    # High-level test:
    lih = Atoms('LiH')
    lih[1].y += 1.64
    lih.center(vacuum=3)

    pos = lih.cell.sum(axis=0)
    print(pos)
    pc = PointChargePotential([-1.0], [pos])
    lih.calc = GPAW(mode='fd',
                    external=pc,
                    mixer=Mixer(0.8, 5, 20.0),
                    xc='oldLDA',
                    poissonsolver=PoissonSolver(),
                    convergence=dict(density=3e-6),
                    txt='lih-pc.txt')
    f1 = lih.get_forces()
    fpc1 = pc.get_forces(lih.calc)
    print(fpc1)
    print(fpc1 + f1.sum(0))

    f2 = [[numeric_force(lih, a, v) for v in range(3)] for a in range(2)]
    print(f1)
    print(f1 - f2)
    err = abs(f1 - f2).max()
    print('err', err)
    assert err < 2e-3, err

    x = 0.0001
    for v in range(3):
        pos[v] += x
        pc.set_positions([pos])
        ep = lih.get_potential_energy()
        pos[v] -= 2 * x
        pc.set_positions([pos])
        em = lih.get_potential_energy()
        pos[v] += x
        pc.set_positions([pos])
        error = (em - ep) / (2 * x) - fpc1[0, v]
        print(v, error)
        assert abs(error) < 0.006



import numpy as np

from gpaw.matrix import Matrix, create_distribution


class MatrixInFile:
    def __init__(self, M, N, dtype, data, dist):
        self.shape = (M, N)
        self.dtype = dtype
        self.array = data  # pointer to data in a file
        self.dist = create_distribution(M, N, *dist)


class ArrayWaveFunctions:
    def __init__(self, M, N, dtype, data, dist, collinear):
        self.collinear = collinear
        if not collinear:
            N *= 2
        if data is None or isinstance(data, np.ndarray):
            self.matrix = Matrix(M, N, dtype, data, dist)
            self.in_memory = True
        else:
            self.matrix = MatrixInFile(M, N, dtype, data, dist)
            self.in_memory = False
        self.comm = None
        self.dtype = self.matrix.dtype

    def __len__(self):
        return len(self.matrix)

    def multiply(self, alpha, opa, b, opb, beta, c, symmetric):
        self.matrix.multiply(alpha, opa, b.matrix, opb, beta, c, symmetric)
        if opa == 'N' and self.comm:
            if self.comm.size > 1:
                c.comm = self.comm
                c.state = 'a sum is needed'
            assert opb in 'TC' and b.comm is self.comm

    def matrix_elements(self, other=None, out=None, symmetric=False, cc=False,
                        operator=None, result=None, serial=False):
        if out is None:
            out = Matrix(len(self), len(other or self), dtype=self.dtype,
                         dist=(self.matrix.dist.comm,
                               self.matrix.dist.rows,
                               self.matrix.dist.columns))
        if other is None or isinstance(other, ArrayWaveFunctions):
            assert cc
            if other is None:
                assert symmetric
                operate_and_multiply(self, self.dv, out, operator, result)
            elif not serial:
                assert not symmetric
                operate_and_multiply_not_symmetric(self, self.dv, out,
                                                   other)
            else:
                self.multiply(self.dv, 'N', other, 'C', 0.0, out, symmetric)
        else:
            assert not cc
            P_ani = {a: P_ni for a, P_ni in out.items()}
            other.integrate(self.array, P_ani, self.kpt)
        return out

    def add(self, lfc, coefs):
        lfc.add(self.array, dict(coefs.items()), self.kpt)

    def apply(self, func, out=None):
        out = out or self.new()
        func(self.array, out.array)
        return out

    def __setitem__(self, i, x):
        x.eval(self.matrix)

    def __iadd__(self, other):
        other.eval(self.matrix, 1.0)
        return self

    def eval(self, matrix):
        matrix.array[:] = self.matrix.array

    def read_from_file(self):
        """Read wave functions from file into memory."""
        matrix = Matrix(*self.matrix.shape,
                        dtype=self.dtype, dist=self.matrix.dist)
        # Read band by band to save memory
        rows = matrix.dist.rows
        blocksize = (matrix.shape[0] + rows - 1) // rows
        for myn, psit_G in enumerate(matrix.array):
            n = matrix.dist.comm.rank * blocksize + myn
            if self.comm.rank == 0:
                big_psit_G = self.array[n]
                if big_psit_G.dtype == complex and self.dtype == float:
                    big_psit_G = big_psit_G.view(float)
                elif big_psit_G.dtype == float and self.dtype == complex:
                    big_psit_G = np.asarray(big_psit_G, complex)
            else:
                big_psit_G = None
            self._distribute(big_psit_G, psit_G)
        self.matrix = matrix
        self.in_memory = True


class UniformGridWaveFunctions(ArrayWaveFunctions):
    def __init__(self, nbands, gd, dtype=None, data=None, kpt=None, dist=None,
                 spin=0, collinear=True):
        ngpts = gd.n_c.prod()
        ArrayWaveFunctions.__init__(self, nbands, ngpts, dtype, data, dist,
                                    collinear)

        M = self.matrix

        if data is None:
            M.array = M.array.reshape(-1).reshape(M.dist.shape)

        self.myshape = (M.dist.shape[0],) + tuple(gd.n_c)
        self.gd = gd
        self.dv = gd.dv
        self.kpt = kpt
        self.spin = spin
        self.comm = gd.comm

    @property
    def array(self):
        if self.in_memory:
            return self.matrix.array.reshape(self.myshape)
        else:
            return self.matrix.array

    def _distribute(self, big_psit_R, psit_R):
        self.gd.distribute(big_psit_R, psit_R.reshape(self.gd.n_c))

    def __repr__(self):
        s = ArrayWaveFunctions.__repr__(self).split('(')[1][:-1]
        shape = self.gd.get_size_of_global_array()
        s = 'UniformGridWaveFunctions({}, gpts={}x{}x{})'.format(s, *shape)
        return s

    def new(self, buf=None, dist='inherit', nbands=None):
        if dist == 'inherit':
            dist = self.matrix.dist
        return UniformGridWaveFunctions(nbands or len(self),
                                        self.gd, self.dtype,
                                        buf,
                                        self.kpt, dist,
                                        self.spin)

    def view(self, n1, n2):
        return UniformGridWaveFunctions(n2 - n1, self.gd, self.dtype,
                                        self.array[n1:n2],
                                        self.kpt, None,
                                        self.spin)

    def plot(self):
        import matplotlib.pyplot as plt
        ax = plt.figure().add_subplot(111)
        a, b, c = self.array.shape[1:]
        ax.plot(self.array[0, a // 2, b // 2])
        plt.show()


class PlaneWaveExpansionWaveFunctions(ArrayWaveFunctions):
    def __init__(self, nbands, pd, dtype=None, data=None, kpt=0, dist=None,
                 spin=0, collinear=True):
        ng = ng0 = pd.myng_q[kpt]
        if data is not None:
            assert data.dtype == complex
        if dtype == float:
            ng *= 2
            if isinstance(data, np.ndarray):
                data = data.view(float)

        ArrayWaveFunctions.__init__(self, nbands, ng, dtype, data, dist,
                                    collinear)
        self.pd = pd
        self.gd = pd.gd
        self.comm = pd.gd.comm
        self.dv = pd.gd.dv / pd.gd.N_c.prod()
        self.kpt = kpt
        self.spin = spin
        if collinear:
            self.myshape = (self.matrix.dist.shape[0], ng0)
        else:
            self.myshape = (self.matrix.dist.shape[0], 2, ng0)

    @property
    def array(self):
        if not self.in_memory:
            return self.matrix.array
        elif self.dtype == float:
            return self.matrix.array.view(complex)
        else:
            return self.matrix.array.reshape(self.myshape)

    def _distribute(self, big_psit_G, psit_G):
        if self.collinear:
            if self.dtype == float:
                if big_psit_G is not None:
                    big_psit_G = big_psit_G.view(complex)
                psit_G = psit_G.view(complex)
            psit_G[:] = self.pd.scatter(big_psit_G, self.kpt)
        else:
            psit_sG = psit_G.reshape((2, -1))
            psit_sG[0] = self.pd.scatter(big_psit_G[0], self.kpt)
            psit_sG[1] = self.pd.scatter(big_psit_G[1], self.kpt)

    def matrix_elements(self, other=None, out=None, symmetric=False, cc=False,
                        operator=None, result=None, serial=False):
        if other is None or isinstance(other, ArrayWaveFunctions):
            if out is None:
                out = Matrix(len(self), len(other or self), dtype=self.dtype,
                             dist=(self.matrix.dist.comm,
                                   self.matrix.dist.rows,
                                   self.matrix.dist.columns))
            assert cc
            if other is None:
                assert symmetric
                operate_and_multiply(self, self.dv, out, operator, result)
            elif not serial:
                assert not symmetric
                operate_and_multiply_not_symmetric(self, self.dv, out,
                                                   other)
            elif self.dtype == complex:
                self.matrix.multiply(self.dv, 'N', other.matrix, 'C',
                                     0.0, out, symmetric)
            else:
                self.matrix.multiply(2 * self.dv, 'N', other.matrix, 'T',
                                     0.0, out, symmetric)
                if self.gd.comm.rank == 0:
                    correction = np.outer(self.matrix.array[:, 0],
                                          other.matrix.array[:, 0])
                    if symmetric:
                        out.array -= 0.5 * self.dv * (correction +
                                                      correction.T)
                    else:
                        out.array -= self.dv * correction
        else:
            assert not cc
            P_ani = {a: P_ni for a, P_ni in out.items()}
            other.integrate(self.array, P_ani, self.kpt)
        return out

    def new(self, buf=None, dist='inherit', nbands=None):
        if buf is not None:
            array = self.array
            buf = buf.ravel()[:array.size]
            buf.shape = array.shape
        if dist == 'inherit':
            dist = self.matrix.dist
        return PlaneWaveExpansionWaveFunctions(nbands or len(self),
                                               self.pd, self.dtype,
                                               buf,
                                               self.kpt, dist,
                                               self.spin, self.collinear)

    def view(self, n1, n2):
        return PlaneWaveExpansionWaveFunctions(n2 - n1, self.pd, self.dtype,
                                               self.array[n1:n2],
                                               self.kpt, None,
                                               self.spin, self.collinear)


def operate_and_multiply(psit1, dv, out, operator, psit2):
    if psit1.comm:
        if psit2 is not None:
            assert psit2.comm is psit1.comm
        if psit1.comm.size > 1:
            out.comm = psit1.comm
            out.state = 'a sum is needed'

    comm = psit1.matrix.dist.comm
    N = len(psit1)
    n = (N + comm.size - 1) // comm.size
    mynbands = len(psit1.matrix.array)

    buf1 = psit1.new(nbands=n, dist=None)
    buf2 = psit1.new(nbands=n, dist=None)
    half = comm.size // 2
    psit = psit1.view(0, mynbands)
    if psit2 is not None:
        psit2 = psit2.view(0, mynbands)

    for r in range(half + 1):
        rrequest = None
        srequest = None

        if r < half:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            skip = (comm.size % 2 == 0 and r == half - 1)
            n1 = min(rrank * n, N)
            n2 = min(n1 + n, N)
            if not (skip and comm.rank < half) and n2 > n1:
                rrequest = comm.receive(buf1.array[:n2 - n1], rrank, 11, False)
            if not (skip and comm.rank >= half) and len(psit1.array) > 0:
                srequest = comm.send(psit1.array, srank, 11, False)

        if r == 0:
            if operator:
                operator(psit1.array, psit2.array)
            else:
                psit2 = psit

        if not (comm.size % 2 == 0 and r == half and comm.rank < half):
            m12 = psit2.matrix_elements(psit, symmetric=(r == 0), cc=True,
                                        serial=True)
            n1 = min(((comm.rank - r) % comm.size) * n, N)
            n2 = min(n1 + n, N)
            out.array[:, n1:n2] = m12.array[:, :n2 - n1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit = buf1
        buf1, buf2 = buf2, buf1

    requests = []
    blocks = []
    nrows = (comm.size - 1) // 2
    for row in range(nrows):
        for column in range(comm.size - nrows + row, comm.size):
            if comm.rank == row:
                n1 = min(column * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    requests.append(
                        comm.send(out.array[:, n1:n2].T.conj().copy(),
                                  column, 12, False))
            elif comm.rank == column:
                n1 = min(row * n, N)
                n2 = min(n1 + n, N)
                if mynbands > 0 and n2 > n1:
                    block = np.empty((mynbands, n2 - n1), out.dtype)
                    blocks.append((n1, n2, block))
                    requests.append(comm.receive(block, row, 12, False))

    comm.waitall(requests)
    for n1, n2, block in blocks:
        out.array[:, n1:n2] = block


def operate_and_multiply_not_symmetric(psit1, dv, out, psit2):
    if psit1.comm:
        if psit2 is not None:
            assert psit2.comm is psit1.comm
        if psit1.comm.size > 1:
            out.comm = psit1.comm
            out.state = 'a sum is needed'

    comm = psit1.matrix.dist.comm
    N = len(psit1)
    n = (N + comm.size - 1) // comm.size
    mynbands = len(psit1.matrix.array)

    buf1 = psit1.new(nbands=n, dist=None)
    buf2 = psit1.new(nbands=n, dist=None)

    psit1 = psit1.view(0, mynbands)
    psit = psit2.view(0, mynbands)
    for r in range(comm.size):
        rrequest = None
        srequest = None

        if r < comm.size - 1:
            srank = (comm.rank + r + 1) % comm.size
            rrank = (comm.rank - r - 1) % comm.size
            n1 = min(rrank * n, N)
            n2 = min(n1 + n, N)
            if n2 > n1:
                rrequest = comm.receive(buf1.array[:n2 - n1], rrank, 11, False)
            if len(psit1.array) > 0:
                srequest = comm.send(psit2.array, srank, 11, False)

        m12 = psit1.matrix_elements(psit, cc=True, serial=True)
        n1 = min(((comm.rank - r) % comm.size) * n, N)
        n2 = min(n1 + n, N)
        out.array[:, n1:n2] = m12.array[:, :n2 - n1]

        if rrequest:
            comm.wait(rrequest)
        if srequest:
            comm.wait(srequest)

        psit = buf1
        buf1, buf2 = buf2, buf1



import pytest
from ase.lattice.compounds import L1_2

from gpaw import GPAW, PW


@pytest.mark.later
def test_xc_qna_band(in_tmp_dir):
    """Cu3Au"""
    ecut = 300
    kpts = (2, 2, 2)

    QNA = {'alpha': 2.0,
           'name': 'QNA',
           'orbital_dependent': False,
           'parameters': {'Au': (0.125, 0.1), 'Cu': (0.0795, 0.005)},
           'setup_name': 'PBE',
           'type': 'qna-gga'}

    atoms = L1_2(['Au', 'Cu'], latticeconstant=3.74)

    calc = GPAW(mode=PW(ecut),
                xc=QNA,
                kpts=kpts,
                parallel={'domain': 1},
                txt='gs.txt')

    atoms.calc = calc
    atoms.get_potential_energy()
    eigs = calc.get_eigenvalues(kpt=0)[:24]
    calc.write('gs.gpw')

    calc = GPAW('gs.gpw').fixed_density(
        parallel={'domain': 1},
        kpts=[[-0.25, 0.25, -0.25]])
    eigs_new = calc.get_eigenvalues(kpt=0)[:24]
    for eold, enew in zip(eigs, eigs_new):
        print(eold, enew, eold - enew)
        assert abs(eold - enew) < 27e-6



import multiprocessing as mp
import os
import re
import sys
import time
from collections import defaultdict

import numpy as np
from scipy.optimize import differential_evolution as DE
from ase import Atoms
from ase.data import covalent_radii, atomic_numbers
from ase.units import Bohr, Ha

from gpaw import GPAW, PW, setup_paths, Mixer, ConvergenceError
from gpaw.atom.generator2 import generate  # , DatasetGenerationError
from gpaw.atom.aeatom import AllElectronAtom
from gpaw.atom.atompaw import AtomPAW
from gpaw.setup import create_setup


my_covalent_radii = covalent_radii.copy()
for e in ['Bk', 'Cf', 'Es', 'Fm', 'Md', 'No', 'Lr']:  # missing radii
    my_covalent_radii[atomic_numbers[e]] = 1.7

my_radii = {
    1: 0.41, 2: 1.46, 3: 1.47, 4: 1.08, 5: 0.83, 6: 0.75, 7: 0.67,
    8: 0.68, 9: 0.7, 10: 1.63, 11: 1.7, 12: 1.47, 13: 1.4, 14: 1.18,
    15: 1.07, 16: 1.15, 17: 1.06, 19: 2.09, 20: 1.71, 21: 1.58, 22: 1.44,
    23: 1.32, 24: 1.28, 25: 1.31, 26: 1.28, 27: 1.28, 28: 1.28, 29: 1.29,
    30: 1.28, 31: 1.26, 32: 1.28, 33: 1.26, 34: 1.18, 35: 1.19, 37: 2.24,
    38: 1.94, 39: 1.73, 40: 1.6, 41: 1.45, 42: 1.39, 43: 1.31, 44: 1.32,
    45: 1.34, 46: 1.39, 47: 1.52, 48: 1.53, 49: 1.57, 50: 1.45, 51: 1.43,
    52: 1.45, 53: 1.42, 56: 1.95, 57: 1.89, 58: 1.84, 59: 1.91, 60: 1.88,
    61: 1.61, 62: 1.85, 63: 1.83, 64: 1.82, 65: 1.76, 66: 1.74, 67: 1.73,
    68: 1.72, 69: 1.71, 70: 1.7, 71: 1.69, 72: 1.57, 73: 1.45, 74: 1.39,
    75: 1.34, 76: 1.35, 77: 1.35, 78: 1.39, 79: 1.38, 80: 1.7, 81: 2.27,
    82: 1.81, 83: 1.59, 84: 1.69, 88: 2.06, 89: 1.95, 90: 1.92}


class PAWDataError(Exception):
    """Error in PAW-data generation."""


class DatasetOptimizer:
    tolerances = np.array([0.2,  # radii
                           0.3,  # log. derivs.
                           40,  # iterations
                           1.2 * 2 / 3 * 300 * 0.1**0.25,  # convergence
                           0.0005,  # eggbox error
                           0.05])  # IP

    def __init__(self, symbol='H', nc=False):
        self.old = False

        self.symbol = symbol
        self.nc = nc

        line = Path('start.txt').read_text()
        words = line.split()
        assert words[1] == symbol
        projectors = words[3]
        radii = [float(f) for f in words[5].split(',')]
        r0 = float(words[7].split(',')[1])

        self.Z = atomic_numbers[symbol]
        rc = self.rc = my_radii[self.Z] / Bohr

        # Parse projectors string:
        pattern = r'(-?\d+\.\d)'
        energies = []
        for m in re.finditer(pattern, projectors):
            energies.append(float(projectors[m.start():m.end()]))
        self.projectors = re.sub(pattern, '{:.1f}', projectors)
        self.nenergies = len(energies)

        self.x = energies + radii + [r0]
        self.bounds = ([(-1.0, 4.0)] * self.nenergies +
                       [(rc * 0.7, rc * 1.0) for r in radii] +
                       [(0.3, rc)])

        self.ecut1 = 450.0
        self.ecut2 = 800.0

        setup_paths[:0] = ['.']

        self.logfile = None
        self.tflush = time.time() + 60

    def run(self):
        print(self.symbol, self.rc / Bohr, self.projectors)
        print(self.x)
        print(self.bounds)
        init = 'latinhypercube'
        if 0:  # os.path.isfile('data.csv'):
            n = len(self.x)
            data = self.read()[:15 * n]
            if np.isfinite(data[:, n]).all() and len(data) == 15 * n:
                init = data[:, :n]

        DE(self, self.bounds, init=init, workers=8, updating='deferred')

    def generate(self, fd, projectors, radii, r0, xc,
                 scalar_relativistic=True, tag=None, logderivs=True):

        if projectors[-1].isupper():
            nderiv0 = 5
        else:
            nderiv0 = 2

        type = 'poly'
        if self.nc:
            type = 'nc'

        try:
            gen = generate(self.symbol, projectors, radii, r0, nderiv0,
                           xc, scalar_relativistic, (type, 4), output=fd)
        except np.linalg.LinAlgError:
            raise PAWDataError('LinAlgError')

        if not scalar_relativistic:
            if not gen.check_all():
                raise PAWDataError('dataset check failed')

        if tag is not None:
            gen.make_paw_setup(tag or None).write_xml()

        r = 1.1 * gen.rcmax

        lmax = 2
        if 'f' in projectors:
            lmax = 3

        error = 0.0
        if logderivs:
            for l in range(lmax + 1):
                emin = -1.5
                emax = 2.0
                n0 = gen.number_of_core_states(l)
                if n0 > 0:
                    e0_n = gen.aea.channels[l].e_n
                    emin = max(emin, e0_n[n0 - 1] + 0.1)
                energies = np.linspace(emin, emax, 100)
                de = energies[1] - energies[0]
                ld1 = gen.aea.logarithmic_derivative(l, energies, r)
                ld2 = gen.logarithmic_derivative(l, energies, r)
                error += abs(ld1 - ld2).sum() * de

        return error

    def parameters(self, x):
        energies = x[:self.nenergies]
        radii = x[self.nenergies:-1]
        r0 = x[-1]
        projectors = self.projectors.format(*energies)
        return energies, radii, r0, projectors

    def __call__(self, x):
        id = mp.current_process().name[-1]
        self.setup = 'de' + id
        energies, radii, r0, projectors = self.parameters(x)

        self.logfile = open(f'data-{id}.csv', 'a')

        with open(f'out-{id}.txt', 'w') as fd:
            errors, msg, convenergies, eggenergies, ips = \
                self.test(fd, projectors, radii, r0)
        error = ((errors / self.tolerances)**2).sum()

        if msg:
            print(msg, x, error, errors, convenergies, eggenergies, ips,
                  file=sys.stderr)

        convenergies += [0] * (7 - len(convenergies))

        print(', '.join(repr(number) for number in
                        list(x) + [error] + errors +
                        convenergies + eggenergies + ips),
              file=self.logfile)

        if time.time() > self.tflush:
            self.logfile.flush()
            self.tflush = time.time() + 60

        return error

    def test_old_paw_data(self):
        with open('old.txt', 'w') as fd:
            self._test_old_paw_data(fd)

    def _test_old_paw_data(self, fd):
        area, niter, convenergies = self.convergence(fd, 'paw')
        eggenergies = self.eggbox(fd, 'paw')
        print('RESULTS:',
              ', '.join(repr(number) for number in
                        [area, niter, max(eggenergies)] +
                        convenergies + eggenergies),
              file=fd)

    def test(self, fd, projectors, radii, r0):
        errors = [np.inf] * 6
        energies = []
        eggenergies = [0, 0, 0]
        ip = 0.0
        ip0 = 0.0
        msg = ''

        try:
            if any(r < r0 for r in radii):
                raise PAWDataError('Core radius too large')

            rc = self.rc
            errors[0] = sum(r - rc for r in radii if r > rc)

            error = 0.0
            for kwargs in [dict(xc='PBE', tag=self.setup),
                           dict(xc='PBE', scalar_relativistic=False),
                           dict(xc='LDA', tag=self.setup),
                           dict(xc='PBEsol'),
                           dict(xc='RPBE'),
                           dict(xc='PW91')]:
                error += self.generate(fd, projectors, radii, r0, **kwargs)
            errors[1] = error

            area, niter, energies = self.convergence(fd)
            errors[2] = niter
            errors[3] = area

            eggenergies = self.eggbox(fd)
            errors[4] = max(eggenergies)

            ip, ip0 = self.ip(fd)
            errors[5] = ip - ip0

        except (ConvergenceError, PAWDataError, RuntimeError,
                np.linalg.LinAlgError) as e:
            msg = str(e)

        return errors, msg, energies, eggenergies, [ip, ip0]

    def eggbox(self, fd, setup='de'):
        energies = []
        for h in [0.16, 0.18, 0.2]:
            a0 = 16 * h
            atoms = Atoms(self.symbol, cell=(a0, a0, 2 * a0), pbc=True)
            if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
                M = 999
                mixer = {'mixer': Mixer(0.01, 5)}
            else:
                M = 333
                mixer = {}
            atoms.calc = GPAW(h=h,
                              xc='PBE',
                              symmetry='off',
                              setups=self.setup,
                              maxiter=M,
                              txt=fd,
                              **mixer)
            atoms.positions += h / 2  # start with broken symmetry
            e0 = atoms.get_potential_energy()
            atoms.positions -= h / 6
            e1 = atoms.get_potential_energy()
            atoms.positions -= h / 6
            e2 = atoms.get_potential_energy()
            atoms.positions -= h / 6
            e3 = atoms.get_potential_energy()
            energies.append(np.ptp([e0, e1, e2, e3]))
        # print(energies)
        return energies

    def convergence(self, fd, setup='de'):
        a = 3.0
        atoms = Atoms(self.symbol, cell=(a, a, a), pbc=True)
        if 58 <= self.Z <= 70 or 90 <= self.Z <= 102:
            M = 999
            mixer = {'mixer': Mixer(0.01, 5)}
        else:
            M = 333
            mixer = {}
        atoms.calc = GPAW(mode=PW(1500),
                          xc='PBE',
                          setups=self.setup,
                          symmetry='off',
                          maxiter=M,
                          txt=fd,
                          **mixer)
        e0 = atoms.get_potential_energy()
        energies = [e0]
        iters = atoms.calc.get_number_of_iterations()
        oldfde = None
        area = 0.0

        def f(x):
            return x**0.25

        for ec in range(800, 200, -100):
            atoms.calc.set(mode=PW(ec))
            atoms.calc.set(eigensolver='rmm-diis')
            de = atoms.get_potential_energy() - e0
            energies.append(de)
            # print(ec, de)
            fde = f(abs(de))
            if fde > f(0.1):
                if oldfde is None:
                    return np.inf, iters, energies
                ec0 = ec + (fde - f(0.1)) / (fde - oldfde) * 100
                area += ((ec + 100) - ec0) * (f(0.1) + oldfde) / 2
                break

            if oldfde is not None:
                area += 100 * (fde + oldfde) / 2
            oldfde = fde

        return area, iters, energies

    def ip(self, fd):
        IP, IP0 = ip(self.symbol, fd, self.setup)
        return IP, IP0

    def read(self):
        data = []
        for i in '12345678s':
            try:
                d = np.loadtxt(f'data-{i}.csv', delimiter=',')
            except OSError:
                if i == 's':
                    pass
                if i == '1':
                    data = np.loadtxt('data.csv', delimiter=',')
                    break
            data += d.tolist()
        data = np.array(data)
        return data[data[:, len(self.x)].argsort()]

    def summary(self, N=10):
        n = len(self.x)
        for x in self.read()[:N]:
            print('{:3} {:2} {:9.1f} ({}) ({}, {})'
                  .format(self.Z,
                          self.symbol,
                          x[n],
                          ', '.join('{:4.1f}'.format(e) + s
                                    for e, s
                                    in zip(x[n + 1:n + 7] / self.tolerances,
                                           'rlciex')),
                          ', '.join('{:+.2f}'.format(e)
                                    for e in x[:self.nenergies]),
                          ', '.join('{:.2f}'.format(r)
                                    for r in x[self.nenergies:n])))

    def best(self):
        n = len(self.x)
        a = self.read()[0]
        x = a[:n]
        error = a[n]
        energies, radii, r0, projectors = self.parameters(x)
        if 1:
            if projectors[-1].isupper():
                nderiv0 = 5
            else:
                nderiv0 = 2
            fmt = '{:3} {:2} -P {:31} -r {:20} -0 {},{:.2f} # {:10.3f}'
            print(fmt.format(self.Z,
                             self.symbol,
                             projectors,
                             ','.join('{0:.2f}'.format(r) for r in radii),
                             nderiv0,
                             r0,
                             error))
        if 1 and error != np.inf and error != np.nan:
            self.generate(None, projectors, radii, r0, 'PBE', True, 'a7o',
                          logderivs=False)


def ip(symbol, fd, setup):
    xc = 'LDA'
    aea = AllElectronAtom(symbol, log=fd)
    aea.initialize()
    aea.run()
    aea.refine()
    aea.scalar_relativistic = True
    aea.refine()
    energy = aea.ekin + aea.eH + aea.eZ + aea.exc
    eigs = []
    for l, channel in enumerate(aea.channels):
        n = l + 1
        for e, f in zip(channel.e_n, channel.f_n):
            if f == 0:
                break
            eigs.append((e, n, l))
            n += 1
    e0, n0, l0 = max(eigs)
    aea = AllElectronAtom(symbol, log=fd)
    aea.add(n0, l0, -1)
    aea.initialize()
    aea.run()
    aea.refine()
    aea.scalar_relativistic = True
    aea.refine()
    IP = aea.ekin + aea.eH + aea.eZ + aea.exc - energy
    IP *= Ha

    s = create_setup(symbol, type=setup, xc=xc)
    f_ln = defaultdict(list)
    for l, f in zip(s.l_j, s. f_j):
        if f:
            f_ln[l].append(f)

    f_sln = [[f_ln[l] for l in range(1 + max(f_ln))]]
    calc = AtomPAW(symbol, f_sln, xc=xc, txt=fd, setups=setup)
    energy = calc.results['energy']
    # eps_n = calc.wfs.kpt_u[0].eps_n

    f_sln[0][l0][-1] -= 1
    calc = AtomPAW(symbol, f_sln, xc=xc, charge=1, txt=fd, setups=setup)
    IP2 = calc.results['energy'] - energy
    return IP, IP2


if __name__ == '__main__':
    import argparse
    from pathlib import Path
    parser = argparse.ArgumentParser(usage='python -m gpaw.atom.optimize '
                                     '[options] [folder folder ...]',
                                     description='Optimize PAW data')
    parser.add_argument('-s', '--summary', type=int)
    parser.add_argument('-b', '--best', action='store_true')
    parser.add_argument('-n', '--norm-conserving', action='store_true')
    parser.add_argument('-o', '--old-setups', action='store_true')
    parser.add_argument('folder', nargs='*')
    args = parser.parse_args()
    folders = [Path(folder) for folder in args.folder or ['.']]
    home = Path.cwd()
    for folder in folders:
        try:
            os.chdir(folder)
            symbol = Path.cwd().name
            do = DatasetOptimizer(symbol)
            if args.summary:
                do.summary(args.summary)
            elif args.old_setups:
                do.test_old_paw_data()
            elif args.best:
                do.best()
            else:
                do.run()
        finally:
            os.chdir(home)



"""Basic test of elph/supercell/

"""
import numpy as np
import pytest

from gpaw.elph import Supercell
from gpaw.mpi import world

SUPERCELL = (2, 1, 1)

g00 = np.array([[[[-1.16418518e-06, 1.26395317e-02],
                  [1.26395317e-02, 4.05925161e-02]],
                 [[3.48680871e-02, 1.26402319e-02],
                  [1.37904135e-02, 5.31904081e-07]]],
                [[[3.48680871e-02, 1.37904135e-02],
                  [1.26402319e-02, 5.31904081e-07]],
                 [[3.59539410e-07, -1.37902297e-02],
                  [-1.37902297e-02, -4.05933006e-02]]]])


@pytest.mark.skipif(world.size > 2,
                    reason='world.size > 2')
@pytest.mark.elph
def test_supercell(module_tmp_path, supercell_cache):
    # Generate supercell_cache
    supercell_cache

    # read supercell matrix
    g_xsNNMM, basis_info = Supercell.load_supercell_matrix()
    assert g_xsNNMM.shape == (6, 1, 2, 2, 2, 2)
    print(g_xsNNMM[0, 0])
    assert g_xsNNMM[0, 0] == pytest.approx(g00, rel=1e-2)



"""Test the Heisenberg model based methodology of the response code."""

import numpy as np
import pytest

from gpaw.response.heisenberg import (calculate_fm_magnon_energies,
                                      calculate_single_site_magnon_energies)


@pytest.mark.ci
def test_single_site_magnons():
    """Check the single site magnon dispersion functionality."""
    rng = get_rng()
    # ---------- Inputs ---------- #

    # Magnetic moment
    mm = 1.
    # q-point grid
    nq = 11
    q_qc = get_randomized_qpoints(nq, rng)

    # Random J_q, with J=0 at q=0
    J_q = rng.rand(q_qc.shape[0])
    J_q[list(q_qc[:, 2]).index(0.)] = 0.

    # Cosine J_qD with different spin wave stiffnesses D
    D_D = np.linspace(400., 800., 5)
    J_qD = np.outer(np.cos(q_qc[:, 2]), D_D)

    # ---------- Script ---------- #

    # Calculate magnon energies
    E_q = calculate_single_site_magnon_energies(J_q, q_qc, mm)
    E_qD = calculate_single_site_magnon_energies(J_qD, q_qc, mm)

    # Check dimensions of arrays
    assert E_q.shape == (q_qc.shape[0],)
    assert E_qD.shape == J_qD.shape

    # Check versus formulas
    assert np.allclose(E_q, -2. / mm * J_q)  # Remember: J(0) = 0
    assert np.allclose(E_qD, 2. / mm * D_D[np.newaxis, :]
                       * (1. - np.cos(q_qc[:, 2]))[:, np.newaxis])


@pytest.mark.ci
def test_single_site_magnons_consistency():
    """Check that the generalized magnon dispersion calculation is consistent
    for a single site system with the simple analytical formula valid in that
    case."""
    rng = get_rng()
    # ---------- Inputs ---------- #

    # Magnetic moment
    mm = 1.
    # q-point grid
    nq = 11
    q_qc = get_randomized_qpoints(nq, rng)

    # Random isotropic exchange constants
    nJsamples = 6  # sample some different random combinations
    J_qx = rng.rand(q_qc.shape[0], nJsamples)

    # ---------- Script ---------- #

    # Calculate assuming a single site
    E_qx = calculate_single_site_magnon_energies(J_qx, q_qc, mm)

    # Calcualte using generalized functionality
    E_qnx = calculate_fm_magnon_energies(J_qx[:, np.newaxis, np.newaxis, :],
                                         q_qc, mm * np.ones((1, nJsamples)))

    # Test self-consistency
    assert E_qnx.shape[0] == E_qx.shape[0]
    assert E_qnx.shape[1] == 1
    assert E_qnx.shape[-1] == E_qx.shape[-1]
    assert np.allclose(E_qnx[:, 0, :], E_qx, atol=1e-8)


@pytest.mark.ci
def test_fm_random_magnons():
    """Check that the functionality to calculate the magnon dispersion of a
    ferromagnetic system with multiple sites works for a randomized system with
    three sites."""
    rng = get_rng()
    # ---------- Inputs ---------- #

    # Magnetic moments
    nsites = 3
    mm_a = 5. * rng.rand(nsites)
    # q-point grid
    nq = 11
    q_qc = get_randomized_qpoints(nq, rng)

    # Random isotropic exchange constants
    J_qab = 1.j * rng.rand(q_qc.shape[0], nsites, nsites)
    J_qab += rng.rand(q_qc.shape[0], nsites, nsites)
    # Take the Hermitian part of random tensor
    J_qab = (J_qab + np.transpose(np.conjugate(J_qab), (0, 2, 1))) / 2.
    # The q=0 component should furthermore be real
    J_qab[list(q_qc[:, 2]).index(0.)].imag = 0.

    # ---------- Script ---------- #

    # Calculate magnon energies
    E_qn = calculate_fm_magnon_energies(J_qab, q_qc, mm_a)
    E_qn = np.sort(E_qn, axis=1)  # Make sure the eigenvalues are sorted

    # Calculate the magnon energies manually
    mm_inv_ab = 2. / np.sqrt(np.outer(mm_a, mm_a))
    J0_ab = np.diag(np.sum(J_qab[list(q_qc[:, 2]).index(0.)], axis=1))
    H_qab = mm_inv_ab[np.newaxis] * (J0_ab[np.newaxis] - J_qab)
    test_E_qn, _ = np.linalg.eig(H_qab)

    assert E_qn.shape == (q_qc.shape[0], nsites)
    assert np.allclose(test_E_qn.imag, 0.)
    assert np.allclose(E_qn, np.sort(test_E_qn.real, axis=1))


@pytest.mark.ci
def test_fm_vectorized_magnons():
    """Check that the functionality to calculate the magnon dispersion of a
    ferromagnetic system with multiple sites works when supplying multiple
    sets of parameters for the same two-site systems."""
    rng = get_rng()
    # ---------- Inputs ---------- #

    # Magnetic moments
    nsites = 2
    nmagmoms = 4  # Test the same J_qab, but with different site magnetizations
    mm_ax = 5. * rng.rand(nsites, nmagmoms)
    # q-point grid
    nq = 11
    q_qc = get_randomized_qpoints(nq, rng)

    # Use a fixed structure for J_qab with known eigenvalues
    cos_q = np.cos(q_qc[:, 2])
    sin_q = np.sin(q_qc[:, 2])
    J_qab = np.empty((nq, nsites, nsites), dtype=complex)
    J_qab[:, 0, 0] = cos_q
    J_qab[:, 0, 1] = 1. + 1.j * sin_q
    J_qab[:, 1, 0] = 1. - 1.j * sin_q
    J_qab[:, 1, 1] = 2. * cos_q

    # Test different energy scales for the exchange interactions
    nJscales = 6
    Jscale_y = 800. * rng.rand(nJscales)

    # Combine different magnetic moments and scale for the exchange
    J_qabxy = np.empty(J_qab.shape + (nmagmoms, nJscales,), dtype=complex)
    J_qabxy[:] = np.tensordot(J_qab, Jscale_y,
                              axes=((), ()))[..., np.newaxis, :]
    mm_axy = np.moveaxis(np.tile(mm_ax, (nJscales, 1, 1)), 0, -1)

    # ---------- Script ---------- #

    # Calculate magnon energies
    E_qnxy = calculate_fm_magnon_energies(J_qabxy, q_qc, mm_axy)
    E_qnxy = np.sort(E_qnxy, axis=1)  # Make sure the eigenvalues are sorted

    # Calculate magnon energies analytically
    H_diag1_qxy = np.sqrt(mm_axy[1][np.newaxis] / mm_axy[0][np.newaxis])\
        * (2. - cos_q[:, np.newaxis, np.newaxis])
    H_diag2_qxy = np.sqrt(mm_axy[0][np.newaxis] / mm_axy[1][np.newaxis])\
        * (3. - 2. * cos_q[:, np.newaxis, np.newaxis])
    H_diag_avg_qxy = (H_diag1_qxy + H_diag2_qxy) / 2.
    H_diag_diff_qxy = (H_diag1_qxy - H_diag2_qxy) / 2.
    pm_n = np.array([-1., 1.])
    E_test_qnxy = H_diag_avg_qxy[:, np.newaxis]\
        + pm_n[np.newaxis, :, np.newaxis, np.newaxis]\
        * np.sqrt(H_diag_diff_qxy[:, np.newaxis]**2.
                  + (1 + sin_q[:, np.newaxis, np.newaxis, np.newaxis]**2.))
    E_test_qnxy *= 2. / np.sqrt(np.prod(mm_axy, axis=0))
    E_test_qnxy *= Jscale_y

    assert np.allclose(E_qnxy, E_test_qnxy)


# ---------- Test functionality ---------- #


def get_randomized_qpoints(nq, rng):
    """Make a simple, but shuffled, q-point array."""
    q_qc = np.zeros((nq, 3), dtype=float)
    q_qc[:, 2] = np.linspace(0., np.pi, nq)
    rng.shuffle(q_qc[:, 2])

    return q_qc


def get_rng():
    """Choose a specific random seed to make the tests reproducible."""
    rng = np.random.RandomState(23)

    return rng



import pytest
from gpaw.utilities import compiled_with_libvdwxc
import numpy as np
from gpaw.grid_descriptor import GridDescriptor
from gpaw.xc.libvdwxc import vdw_df, vdw_df2, vdw_df_cx, \
    vdw_optPBE, vdw_optB88, vdw_C09, vdw_beef, \
    libvdwxc_has_mpi, libvdwxc_has_pfft

pytestmark = pytest.mark.skipif(not compiled_with_libvdwxc(),
                                reason='not compiled_with_libvdwxc()')


def test_vdw_libvdwxc_functionals():

    # This test verifies that the results returned by the van der Waals
    # functionals implemented in libvdwxc do not change.

    N_c = np.array([23, 10, 6])
    gd = GridDescriptor(N_c, N_c * 0.2, pbc_c=(1, 0, 1))

    n_sg = gd.zeros(1)
    nG_sg = gd.collect(n_sg)
    if gd.comm.rank == 0:
        gen = np.random.RandomState(0)
        nG_sg[:] = gen.rand(*nG_sg.shape)
    gd.distribute(nG_sg, n_sg)

    for mode in ['serial', 'mpi', 'pfft']:
        if mode == 'serial' and gd.comm.size > 1:
            continue
        if mode == 'mpi' and not libvdwxc_has_mpi():
            continue
        if mode == 'pfft' and not libvdwxc_has_pfft():
            continue

        errs = []

        def test(vdwxcclass, Eref=np.nan, nvref=np.nan):
            print('')
            xc = vdwxcclass(mode=mode)
            xc.initialize_backend(gd)
            if gd.comm.rank == 0:
                print(xc.libvdwxc.tostring())
            v_sg = gd.zeros(1)
            E = xc.calculate(gd, n_sg, v_sg)
            nv = gd.integrate(n_sg * v_sg, global_integral=True)
            nv = float(nv)  # Comes out as an array due to spin axis

            Eerr = abs(E - Eref)
            nverr = abs(nv - nvref)
            errs.append((vdwxcclass.__name__, Eerr, nverr))

            if gd.comm.rank == 0:
                name = xc.name
                print(name)
                print('=' * len(name))
                print('E  = %19.16f vs ref = %19.16f :: err = %10.6e'
                      % (E, Eref, Eerr))
                print('nv = %19.16f vs ref = %19.16f :: err = %10.6e'
                      % (nv, nvref, nverr))
                print()
            gd.comm.barrier()

            print('Update:')
            print('    test({}, {!r}, {!r})'.format(vdwxcclass.__name__,
                                                    E, nv))
        test(vdw_df, -3.8730338590248383, -4.905182929615311)
        test(vdw_df2, -3.9017518972499317, -4.933889152385742)
        test(vdw_df_cx, -3.7262108992876644, -4.760433536500078)
        test(vdw_optPBE, -3.7954301587466506, -4.834460613766266)
        test(vdw_optB88, -3.8486341203104613, -4.879005564922708)
        test(vdw_C09, -3.7071083039260464, -4.746114441237086)
        test(vdw_beef, -3.8926148228224444, -4.961101745896925)

        if any(err[1] > 1e-14 or err[2] > 1e-14 for err in errs):
            # Try old values (compatibility)
            del errs[:]

            test(vdw_df, -3.8730338473027368, -4.905182296422721)
            test(vdw_df2, -3.9017516508476211, -4.933888350723616)
            test(vdw_df_cx, -3.7262108875655624, -4.760432903307487)
            test(vdw_optPBE, -3.7954301470245491, -4.834459980573675)
            test(vdw_optB88, -3.8486341085883597, -4.879004931730118)
            test(vdw_C09, -3.7071082922039449, -4.746113808044496)
            test(vdw_beef, -3.8926145764201334, -4.9611009442348015)

            for name, Eerr, nverr in errs:
                assert Eerr < 1e-14 and nverr < 1e-14, (name, Eerr, nverr)



import numpy as np
from gpaw.blacs import BlacsDescriptor, BlacsGrid, Redistributor


class Blocks1D:
    def __init__(self, blockcomm, N):
        self.blockcomm = blockcomm
        self.N = N  # Global number of points

        self.blocksize = (N + blockcomm.size - 1) // blockcomm.size
        self.a = min(blockcomm.rank * self.blocksize, N)
        self.b = min(self.a + self.blocksize, N)
        self.nlocal = self.b - self.a

        self.myslice = slice(self.a, self.b)

    def all_gather(self, in_myix):
        """All-gather array where the first dimension is block distributed.

        Here, myi is understood as the distributed index, whereas x are the
        remaining (global) dimensions."""
        # Set up buffers
        buf_myix = self.local_communication_buffer(in_myix)
        buf_ix = self.global_communication_buffer(in_myix)

        # Excecute all-gather
        self.blockcomm.all_gather(buf_myix, buf_ix)
        out_ix = buf_ix[:self.N]

        return out_ix

    def gather(self, in_myix, root=0):
        """Gather array to root where the first dimension is block distributed.
        """
        assert root in range(self.blockcomm.size)

        # Set up buffers
        buf_myix = self.local_communication_buffer(in_myix)
        if self.blockcomm.rank == root:
            buf_ix = self.global_communication_buffer(in_myix)
        else:
            buf_ix = None

        # Excecute gather
        self.blockcomm.gather(buf_myix, root, buf_ix)
        if self.blockcomm.rank == root:
            out_ix = buf_ix[:self.N]
        else:
            out_ix = None

        return out_ix

    def local_communication_buffer(self, in_myix):
        """Set up local communication buffer."""
        if in_myix.shape[0] == self.blocksize and in_myix.flags.contiguous:
            buf_myix = in_myix  # Use input array as communication buffer
        else:
            assert in_myix.shape[0] == self.nlocal
            buf_myix = np.empty(
                (self.blocksize,) + in_myix.shape[1:], in_myix.dtype)
            buf_myix[:self.nlocal] = in_myix

        return buf_myix

    def global_communication_buffer(self, in_myix):
        """Set up global communication buffer."""
        buf_ix = np.empty(
            (self.blockcomm.size * self.blocksize,) + in_myix.shape[1:],
            dtype=in_myix.dtype)
        return buf_ix

    def find_global_index(self, i):
        """Find rank and local index of the global index i"""
        rank = i // self.blocksize
        li = i % self.blocksize

        return rank, li


def block_partition(comm, nblocks):
    r"""Partition the communicator into a 2D array with horizontal
    and vertical communication.

         Communication between blocks (blockcomm)
    <----------------------------------------------->
     _______________________________________________
    |     |     |     |     |     |     |     |     | ⋀
    |  0  |  1  |  2  |  3  |  4  |  5  |  6  |  7  | |
    |_____|_____|_____|_____|_____|_____|_____|_____| |
    |     |     |     |     |     |     |     |     | | Communication inside
    |  8  |  9  | 10  | 11  | 12  | 13  | 14  | 15  | | blocks
    |_____|_____|_____|_____|_____|_____|_____|_____| | (intrablockcomm)
    |     |     |     |     |     |     |     |     | |
    | 16  | 17  | 18  | 19  | 20  | 21  | 22  | 23  | |
    |_____|_____|_____|_____|_____|_____|_____|_____| ⋁
    
    """
    if nblocks == 'max':
        # Maximize the number of blocks
        nblocks = comm.size
    assert isinstance(nblocks, int)
    assert nblocks > 0 and nblocks <= comm.size, comm.size
    assert comm.size % nblocks == 0, comm.size

    # Communicator between different blocks
    if nblocks == comm.size:
        blockcomm = comm
    else:
        rank1 = comm.rank // nblocks * nblocks
        rank2 = rank1 + nblocks
        blockcomm = comm.new_communicator(range(rank1, rank2))

    # Communicator inside each block
    ranks = range(comm.rank % nblocks, comm.size, nblocks)
    if nblocks == 1:
        assert len(ranks) == comm.size
        intrablockcomm = comm
    else:
        intrablockcomm = comm.new_communicator(ranks)

    assert blockcomm.size * intrablockcomm.size == comm.size

    return blockcomm, intrablockcomm


class PlaneWaveBlockDistributor:
    """Functionality to shuffle block distribution of pair functions
    in the plane wave basis."""

    def __init__(self, world, blockcomm, intrablockcomm):
        self.world = world
        self.blockcomm = blockcomm
        self.intrablockcomm = intrablockcomm

    @property
    def fully_block_distributed(self):
        return self.world.compare(self.blockcomm) == 'ident'

    def new_distributor(self, *, nblocks):
        """Set up a new PlaneWaveBlockDistributor."""
        world = self.world
        blockcomm, intrablockcomm = block_partition(comm=world,
                                                    nblocks=nblocks)
        blockdist = PlaneWaveBlockDistributor(world, blockcomm, intrablockcomm)

        return blockdist

    def _redistribute(self, in_wGG, nw):
        """Redistribute array.

        Switch between two kinds of parallel distributions:

        1) parallel over G-vectors (second dimension of in_wGG)
        2) parallel over frequency (first dimension of in_wGG)

        Returns new array using the memory in the 1-d array out_x.
        """

        comm = self.blockcomm

        if comm.size == 1:
            return in_wGG

        mynw = (nw + comm.size - 1) // comm.size
        nG = in_wGG.shape[2]
        mynG = (nG + comm.size - 1) // comm.size

        bg1 = BlacsGrid(comm, comm.size, 1)
        bg2 = BlacsGrid(comm, 1, comm.size)
        md1 = BlacsDescriptor(bg1, nw, nG**2, mynw, nG**2)
        md2 = BlacsDescriptor(bg2, nw, nG**2, nw, mynG * nG)

        if len(in_wGG) == nw:
            mdin = md2
            mdout = md1
        else:
            mdin = md1
            mdout = md2

        r = Redistributor(comm, mdin, mdout)

        # mdout.shape[1] is always divisible by nG because
        # every block starts at a multiple of nG, and the last block
        # ends at nG² which of course also is divisible.  Nevertheless:
        assert mdout.shape[1] % nG == 0
        # (If it were not divisible, we would "lose" some numbers and the
        #  redistribution would be corrupted.)

        inbuf = in_wGG.reshape(mdin.shape)
        # numpy.reshape does not *guarantee* that the reshaped view will
        # be contiguous. To support redistribution of input arrays with an
        # arbitrary allocation layout, we make sure that the corresponding
        # input BLACS buffer in contiguous
        inbuf = np.ascontiguousarray(inbuf)

        outbuf = np.empty(mdout.shape, complex)

        r.redistribute(inbuf, outbuf)

        outshape = (mdout.shape[0], mdout.shape[1] // nG, nG)
        out_wGG = outbuf.reshape(outshape)
        assert out_wGG.flags.contiguous  # Since mdout.shape[1] % nG == 0

        return out_wGG

    def has_distribution(self, array, nw, distribution):
        """Check if array 'array' has distribution 'distribution'."""
        if distribution not in ['wGG', 'WgG', 'zGG', 'ZgG']:
            raise ValueError(f'Invalid dist_type: {distribution}')

        comm = self.blockcomm
        if comm.size == 1:
            # In serial, all distributions are equivalent
            return True

        # At the moment, only wGG and WgG distributions are supported. zGG and
        # ZgG are complex frequency aliases for wGG and WgG respectively
        assert len(array.shape) == 3
        nG = array.shape[2]
        if array.shape[1] < array.shape[2]:
            # Looks like array is WgG/ZgG distributed
            gblocks = Blocks1D(comm, nG)
            assert array.shape == (nw, gblocks.nlocal, nG)
            return distribution in ['WgG', 'ZgG']
        else:
            # Looks like array is wGG/zGG distributed
            wblocks = Blocks1D(comm, nw)
            assert array.shape == (wblocks.nlocal, nG, nG)
            return distribution in ['wGG', 'zGG']
        
    def distribute_as(self, array, nw, distribution):
        """Redistribute array.

        Switch between two kinds of parallel distributions:

        1) parallel over G-vectors (distribution in ['WgG', 'ZgG'])
        2) parallel over frequency (distribution in ['wGG', 'zGG'])
        """
        if self.has_distribution(array, nw, distribution):
            # If the array already has the requested distribution, do nothing
            return array
        else:
            return self._redistribute(array, nw)
    
    def distribute_frequencies(self, in_wGG, nw):
        """Distribute frequencies to all cores."""

        world = self.world
        comm = self.blockcomm

        if world.size == 1:
            return in_wGG

        mynw = (nw + world.size - 1) // world.size
        nG = in_wGG.shape[2]
        mynG = (nG + comm.size - 1) // comm.size

        wa = min(world.rank * mynw, nw)
        wb = min(wa + mynw, nw)

        if self.blockcomm.size == 1:
            return in_wGG[wa:wb].copy()

        if self.intrablockcomm.rank == 0:
            bg1 = BlacsGrid(comm, 1, comm.size)
            in_wGG = in_wGG.reshape((nw, -1))
        else:
            bg1 = BlacsGrid(None, 1, 1)
            # bg1 = DryRunBlacsGrid(mpi.serial_comm, 1, 1)
            in_wGG = np.zeros((0, 0), complex)
        md1 = BlacsDescriptor(bg1, nw, nG**2, nw, mynG * nG)

        bg2 = BlacsGrid(world, world.size, 1)
        md2 = BlacsDescriptor(bg2, nw, nG**2, mynw, nG**2)

        r = Redistributor(world, md1, md2)
        shape = (wb - wa, nG, nG)
        out_wGG = np.empty(shape, complex)
        r.redistribute(in_wGG, out_wGG.reshape((wb - wa, nG**2)))

        return out_wGG



"""Example from the gpaw-dev ML.


    https://listserv.fysik.dtu.dk/pipermail/gpaw-users/2016-January/003864.html
"""
from ase import Atoms
from gpaw.symmetry import atoms2symmetry


def test_symmetry_check():
    a = 3.9
    atoms = Atoms('Ge2Si4',
                  cell=[[a, 0, 0],
                        [0, a, 0],
                        [-a / 2, -a / 2, a * 3 / 2**0.5]],
                  scaled_positions=[[2 / 3, 2 / 3, 1 / 3],
                                    [11 / 12, 5 / 12, 5 / 6],
                                    [0, 0, 0],
                                    [1 / 4, 3 / 4, 1 / 2],
                                    [1 / 3, 1 / 3, 2 / 3],
                                    [7 / 12, 1 / 12, 1 / 6]],
                  pbc=True)
    sym = atoms2symmetry(atoms)
    ok = sym.check_grid((20, 20, 40))
    assert not ok



from math import pi

import numpy as np

from gpaw.core import UGDesc
from gpaw.new import zips, spinsum
from gpaw.new.pot_calc import PotentialCalculator


class FDPotentialCalculator(PotentialCalculator):
    def __init__(self,
                 wf_grid: UGDesc,
                 fine_grid: UGDesc,
                 setups,
                 xc,
                 poisson_solver,
                 *,
                 fracpos_ac,
                 atomdist,
                 interpolation_stencil_range=3,
                 xp=np):
        self.fine_grid = fine_grid
        self.grid = wf_grid

        self.vbar_ar = setups.create_local_potentials(fine_grid, fracpos_ac,
                                                      atomdist, xp=xp)
        self.ghat_aLr = setups.create_compensation_charges(fine_grid,
                                                           fracpos_ac,
                                                           atomdist,
                                                           xp=xp)

        self.vbar_r = fine_grid.empty(xp=xp)
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)

        n = interpolation_stencil_range
        self.interpolation_stencil_range = n
        self._interpolate = wf_grid.transformer(fine_grid, n, xp=xp)
        self._restrict = fine_grid.transformer(wf_grid, n, xp=xp)

        super().__init__(xc, poisson_solver, setups,
                         fracpos_ac=fracpos_ac)

    def __str__(self):
        txt = super().__str__()
        degree = self.interpolation_stencil_range * 2 - 1
        name = ['linear', 'cubic', 'quintic', 'heptic'][degree // 2]
        txt += (f'interpolation: tri-{name}' +
                f' # {degree}. degree polynomial\n')
        return txt

    def interpolate(self, a_xR, a_xr=None):
        return self._interpolate(a_xR, a_xr)

    def restrict(self, a_xr, a_xR=None):
        return self._restrict(a_xr, a_xR)

    def calculate_charges(self, vHt_r):
        return self.ghat_aLr.integrate(vHt_r)

    def calculate_non_selfconsistent_exc(self, xc, nt_sR, taut_sR):
        nt_sr, _, _ = self._interpolate_density(nt_sR)
        if taut_sR is not None:
            taut_sr = self.interpolate(taut_sR)
        else:
            taut_sr = None
        e_xc, _, _ = xc.calculate(nt_sr, taut_sr)
        return e_xc

    def _interpolate_density(self, nt_sR):
        nt_sr = self.interpolate(nt_sR)
        if not nt_sR.desc.pbc_c.all():
            Nt1_s = nt_sR.integrate()
            Nt2_s = nt_sr.integrate()
            for Nt1, Nt2, nt_r in zips(Nt1_s, Nt2_s, nt_sr):
                if Nt2 > 1e-14:
                    nt_r.data *= Nt1 / Nt2
        return nt_sr, None, None

    def calculate_pseudo_potential(self, density, ibzwfs, vHt_r):
        nt_sr, _, _ = self._interpolate_density(density.nt_sR)
        grid2 = nt_sr.desc

        if density.taut_sR is not None:
            taut_sr = self.interpolate(density.taut_sR)
        else:
            taut_sr = None

        e_xc, vxct_sr, dedtaut_sr = self.xc.calculate(nt_sr, taut_sr)

        charge_r = grid2.empty()
        charge_r.data[:] = nt_sr.data[:density.ndensities].sum(axis=0)
        e_zero = self.vbar_r.integrate(charge_r)

        ccc_aL = density.calculate_compensation_charge_coefficients()

        # Normalize: (LCAO basis functions may extend outside box)
        comp_charge = (4 * pi)**0.5 * sum(ccc_L[0]
                                          for ccc_L in ccc_aL.values())
        comp_charge = ccc_aL.layout.atomdist.comm.sum(comp_charge)
        pseudo_charge = charge_r.integrate()
        charge_r.data *= -(comp_charge + density.charge) / pseudo_charge

        self.ghat_aLr.add_to(charge_r, ccc_aL)

        if vHt_r is None:
            vHt_r = grid2.zeros()
        self.poisson_solver.solve(vHt_r, charge_r)
        e_coulomb = 0.5 * vHt_r.integrate(charge_r)

        vt_sr = vxct_sr
        vt_sr.data += vHt_r.data + self.vbar_r.data
        vt_sR = self.restrict(vt_sr)

        e_external = 0.0

        return {'coulomb': e_coulomb,
                'zero': e_zero,
                'xc': e_xc,
                'external': e_external}, vt_sR, dedtaut_sr, vHt_r

    def move(self, fracpos_ac, atomdist):
        self.ghat_aLr.move(fracpos_ac, atomdist)
        self.vbar_ar.move(fracpos_ac, atomdist)
        self.vbar_ar.to_uniform_grid(out=self.vbar_r)

    def force_contributions(self, state):
        density = state.density
        potential = state.potential
        nt_R = spinsum(density.nt_sR)
        vt_R = spinsum(potential.vt_sR, mean=True)
        dedtaut_sR = potential.dedtaut_sR
        if dedtaut_sR is not None:
            dedtaut_R = spinsum(dedtaut_sR, mean=True)
            Ftauct_av = state.density.tauct_aX.derivative(dedtaut_R)
        else:
            Ftauct_av = None

        nt_r = self.interpolate(nt_R)
        if not nt_r.desc.pbc_c.all():
            scale = nt_R.integrate() / nt_r.integrate()
            nt_r.data *= scale

        return (self.ghat_aLr.derivative(state.potential.vHt_x),
                state.density.nct_aX.derivative(vt_R),
                Ftauct_av,
                self.vbar_ar.derivative(nt_r))
